# /*##########################################################################
#
# Copyright (c) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
This module contains an :class:`SafeFileSystemModel`.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "22/11/2017"

import sys
import os.path
import logging
import weakref

from silx.gui import qt
from .SafeFileIconProvider import SafeFileIconProvider

_logger = logging.getLogger(__name__)


class _Item(object):

    def __init__(self, fileInfo):
        self.__fileInfo = fileInfo
        self.__parent = None
        self.__children = None
        self.__absolutePath = None

    def isDrive(self):
        if sys.platform == "win32":
            return self.parent().parent() is None
        else:
            return False

    def isRoot(self):
        return self.parent() is None

    def isFile(self):
        """
        Returns true if the path is a file.

        It avoid to access to the `Qt.QFileInfo` in case the file is a drive.
        """
        if self.isDrive():
            return False
        return self.__fileInfo.isFile()

    def isDir(self):
        """
        Returns true if the path is a directory.

        The default `qt.QFileInfo.isDir` can freeze the file system with
        network drives. This function avoid the freeze in case of browsing
        the root.
        """
        if self.isDrive():
            # A drive is a directory, we don't have to synchronize the
            # drive to know that
            return True
        return self.__fileInfo.isDir()

    def absoluteFilePath(self):
        """
        Returns an absolute path including the file name.

        This function uses in most cases the default
        `qt.QFileInfo.absoluteFilePath`. But it is known to freeze the file
        system with network drives.

        This function uses `qt.QFileInfo.filePath` in case of root drives, to
        avoid this kind of issues. In case of drive, the result is the same,
        while the file path is already absolute.

        :rtype: str
        """
        if self.__absolutePath is None:
            if self.isRoot():
                path = ""
            elif self.isDrive():
                path = self.__fileInfo.filePath()
            else:
                path = os.path.join(self.parent().absoluteFilePath(), self.__fileInfo.fileName())
                if path == "":
                    return "/"
            self.__absolutePath = path
        return self.__absolutePath

    def child(self):
        self.populate()
        return self.__children

    def childAt(self, position):
        self.populate()
        return self.__children[position]

    def childCount(self):
        self.populate()
        return len(self.__children)

    def indexOf(self, item):
        self.populate()
        return self.__children.index(item)

    def parent(self):
        parent = self.__parent
        if parent is None:
            return None
        return parent()

    def filePath(self):
        return self.__fileInfo.filePath()

    def fileName(self):
        if self.isDrive():
            name = self.absoluteFilePath()
            if name[-1] == "/":
                name = name[:-1]
            return name
        return os.path.basename(self.absoluteFilePath())

    def fileInfo(self):
        """
        Returns the Qt file info.

        :rtype: Qt.QFileInfo
        """
        return self.__fileInfo

    def _setParent(self, parent):
        self.__parent = weakref.ref(parent)

    def findChildrenByPath(self, path):
        if path == "":
            return self
        path = path.replace("\\", "/")
        if path[-1] == "/":
            path = path[:-1]
        names = path.split("/")
        caseSensitive = qt.QFSFileEngine(path).caseSensitive()
        count = len(names)
        cursor = self
        for name in names:
            for item in cursor.child():
                if caseSensitive:
                    same = item.fileName() == name
                else:
                    same = item.fileName().lower() == name.lower()
                if same:
                    cursor = item
                    count -= 1
                    break
            else:
                return None
            if count == 0:
                break
        else:
            return None
        return cursor

    def populate(self):
        if self.__children is not None:
            return
        self.__children = []
        if self.isRoot():
            items = qt.QDir.drives()
        else:
            directory = qt.QDir(self.absoluteFilePath())
            filters = qt.QDir.AllEntries | qt.QDir.Hidden | qt.QDir.System
            items = directory.entryInfoList(filters)
        for fileInfo in items:
            i = _Item(fileInfo)
            self.__children.append(i)
            i._setParent(self)


class _RawFileSystemModel(qt.QAbstractItemModel):
    """
    This class implement a file system model and try to avoid freeze. On Qt4,
    :class:`qt.QFileSystemModel` is known to freeze the file system when
    network drives are available.

    To avoid this behaviour, this class does not use
    `qt.QFileInfo.absoluteFilePath` nor `qt.QFileInfo.canonicalPath` to reach
    information on drives.

    This model do not take care of sorting and filtering. This features are
    managed by another model, by composition.

    And because it is the end of life of Qt4, we do not implement asynchronous
    loading of files as it is done by :class:`qt.QFileSystemModel`, nor some
    useful features.
    """

    __directoryLoadedSync = qt.Signal(str)
    """This signal is connected asynchronously to a slot. It allows to
    emit directoryLoaded as an asynchronous signal."""

    directoryLoaded = qt.Signal(str)
    """This signal is emitted when the gatherer thread has finished to load the
    path."""

    rootPathChanged = qt.Signal(str)
    """This signal is emitted whenever the root path has been changed to a
    newPath."""

    NAME_COLUMN = 0
    SIZE_COLUMN = 1
    TYPE_COLUMN = 2
    LAST_MODIFIED_COLUMN = 3

    def __init__(self, parent=None):
        qt.QAbstractItemModel.__init__(self, parent)
        self.__computer = _Item(qt.QFileInfo())
        self.__header = "Name", "Size", "Type", "Last modification"
        self.__currentPath = ""
        self.__iconProvider = SafeFileIconProvider()
        self.__directoryLoadedSync.connect(self.__emitDirectoryLoaded, qt.Qt.QueuedConnection)

    def headerData(self, section, orientation, role=qt.Qt.DisplayRole):
        if orientation == qt.Qt.Horizontal:
            if role == qt.Qt.DisplayRole:
                return self.__header[section]
            if role == qt.Qt.TextAlignmentRole:
                return qt.Qt.AlignRight if section == 1 else qt.Qt.AlignLeft
        return None

    def flags(self, index):
        if not index.isValid():
            return 0
        return qt.Qt.ItemIsEnabled | qt.Qt.ItemIsSelectable

    def columnCount(self, parent=qt.QModelIndex()):
        return len(self.__header)

    def rowCount(self, parent=qt.QModelIndex()):
        item = self.__getItem(parent)
        return item.childCount()

    def data(self, index, role=qt.Qt.DisplayRole):
        if not index.isValid():
            return None

        column = index.column()
        if role in [qt.Qt.DisplayRole, qt.Qt.EditRole]:
            if column == self.NAME_COLUMN:
                return self.__displayName(index)
            elif column == self.SIZE_COLUMN:
                return self.size(index)
            elif column == self.TYPE_COLUMN:
                return self.type(index)
            elif column == self.LAST_MODIFIED_COLUMN:
                return self.lastModified(index)
            else:
                _logger.warning("data: invalid display value column %d", index.column())
        elif role == qt.QFileSystemModel.FilePathRole:
            return self.filePath(index)
        elif role == qt.QFileSystemModel.FileNameRole:
            return self.fileName(index)
        elif role == qt.Qt.DecorationRole:
            if column == self.NAME_COLUMN:
                icon = self.fileIcon(index)
                if icon is None or icon.isNull():
                    if self.isDir(index):
                        self.__iconProvider.icon(qt.QFileIconProvider.Folder)
                    else:
                        self.__iconProvider.icon(qt.QFileIconProvider.File)
                return icon
        elif role == qt.Qt.TextAlignmentRole:
            if column == self.SIZE_COLUMN:
                return qt.Qt.AlignRight
        elif role == qt.QFileSystemModel.FilePermissions:
            return self.permissions(index)

        return None

    def index(self, *args, **kwargs):
        path_api = False
        path_api |= len(args) >= 1 and isinstance(args[0], str)
        path_api |= "path" in kwargs

        if path_api:
            return self.__indexFromPath(*args, **kwargs)
        else:
            return self.__index(*args, **kwargs)

    def __index(self, row, column, parent=qt.QModelIndex()):
        if parent.isValid() and parent.column() != 0:
            return None

        parentItem = self.__getItem(parent)
        item = parentItem.childAt(row)
        return self.createIndex(row, column, item)

    def __indexFromPath(self, path, column=0):
        """
        Uses the index(str) C++ API

        :rtype: qt.QModelIndex
        """
        if path == "":
            return qt.QModelIndex()

        item = self.__computer.findChildrenByPath(path)
        if item is None:
            return qt.QModelIndex()

        return self.createIndex(item.parent().indexOf(item), column, item)

    def parent(self, index):
        if not index.isValid():
            return qt.QModelIndex()

        item = self.__getItem(index)
        if index is None:
            return qt.QModelIndex()

        parent = item.parent()
        if parent is None or parent is self.__computer:
            return qt.QModelIndex()

        return self.createIndex(parent.parent().indexOf(parent), 0, parent)

    def __emitDirectoryLoaded(self, path):
        self.directoryLoaded.emit(path)

    def __emitRootPathChanged(self, path):
        self.rootPathChanged.emit(path)

    def __getItem(self, index):
        if not index.isValid():
            return self.__computer
        item = index.internalPointer()
        return item

    def fileIcon(self, index):
        item = self.__getItem(index)
        if self.__iconProvider is not None:
            fileInfo = item.fileInfo()
            result = self.__iconProvider.icon(fileInfo)
        else:
            style = qt.QApplication.instance().style()
            if item.isRoot():
                result = style.standardIcon(qt.QStyle.SP_ComputerIcon)
            elif item.isDrive():
                result = style.standardIcon(qt.QStyle.SP_DriveHDIcon)
            elif item.isDir():
                result = style.standardIcon(qt.QStyle.SP_DirIcon)
            else:
                result = style.standardIcon(qt.QStyle.SP_FileIcon)
        return result

    def _item(self, index):
        item = self.__getItem(index)
        return item

    def fileInfo(self, index):
        item = self.__getItem(index)
        result = item.fileInfo()
        return result

    def __fileIcon(self, index):
        item = self.__getItem(index)
        result = item.fileName()
        return result

    def __displayName(self, index):
        item = self.__getItem(index)
        result = item.fileName()
        return result

    def fileName(self, index):
        item = self.__getItem(index)
        result = item.fileName()
        return result

    def filePath(self, index):
        item = self.__getItem(index)
        result = item.fileInfo().filePath()
        return result

    def isDir(self, index):
        item = self.__getItem(index)
        result = item.isDir()
        return result

    def lastModified(self, index):
        item = self.__getItem(index)
        result = item.fileInfo().lastModified()
        return result

    def permissions(self, index):
        item = self.__getItem(index)
        result = item.fileInfo().permissions()
        return result

    def size(self, index):
        item = self.__getItem(index)
        result = item.fileInfo().size()
        return result

    def type(self, index):
        item = self.__getItem(index)
        if self.__iconProvider is not None:
            fileInfo = item.fileInfo()
            result = self.__iconProvider.type(fileInfo)
        else:
            if item.isRoot():
                result = "Computer"
            elif item.isDrive():
                result = "Drive"
            elif item.isDir():
                result = "Directory"
            else:
                fileInfo = item.fileInfo()
                result = fileInfo.suffix()
        return result

    # File manipulation

    # bool        remove(const QModelIndex & index) const
    # bool        rmdir(const QModelIndex & index) const
    # QModelIndex mkdir(const QModelIndex & parent, const QString & name)

    # Configuration

    def rootDirectory(self):
        return qt.QDir(self.rootPath())

    def rootPath(self):
        return self.__currentPath

    def setRootPath(self, path):
        if self.__currentPath == path:
            return
        self.__currentPath = path
        item = self.__computer.findChildrenByPath(path)
        self.__emitRootPathChanged(path)
        if item is None or item.parent() is None:
            return qt.QModelIndex()
        index = self.createIndex(item.parent().indexOf(item), 0, item)
        self.__directoryLoadedSync.emit(path)
        return index

    def iconProvider(self):
        # FIXME: invalidate the model
        return self.__iconProvider

    def setIconProvider(self, provider):
        # FIXME: invalidate the model
        self.__iconProvider = provider

    # bool               resolveSymlinks() const
    # void               setResolveSymlinks(bool enable)

    def setNameFilterDisables(self, enable):
        return None

    def nameFilterDisables(self):
        return None

    def myComputer(self, role=qt.Qt.DisplayRole):
        return None

    def setNameFilters(self, filters):
        return

    def nameFilters(self):
        return None

    def filter(self):
        return self.__filters

    def setFilter(self, filters):
        return

    def setReadOnly(self, enable):
        assert(enable is True)

    def isReadOnly(self):
        return False


class SafeFileSystemModel(qt.QSortFilterProxyModel):
    """
    This class implement a file system model and try to avoid freeze. On Qt4,
    :class:`qt.QFileSystemModel` is known to freeze the file system when
    network drives are available.

    To avoid this behaviour, this class does not use
    `qt.QFileInfo.absoluteFilePath` nor `qt.QFileInfo.canonicalPath` to reach
    information on drives.

    And because it is the end of life of Qt4, we do not implement asynchronous
    loading of files as it is done by :class:`qt.QFileSystemModel`, nor some
    useful features.
    """

    def __init__(self, parent=None):
        qt.QSortFilterProxyModel.__init__(self, parent=parent)
        self.__nameFilterDisables = sys.platform == "darwin"
        self.__nameFilters = []
        self.__filters = qt.QDir.AllEntries | qt.QDir.NoDotAndDotDot | qt.QDir.AllDirs
        sourceModel = _RawFileSystemModel(self)
        self.setSourceModel(sourceModel)

    @property
    def directoryLoaded(self):
        return self.sourceModel().directoryLoaded

    @property
    def rootPathChanged(self):
        return self.sourceModel().rootPathChanged

    def index(self, *args, **kwargs):
        path_api = False
        path_api |= len(args) >= 1 and isinstance(args[0], str)
        path_api |= "path" in kwargs

        if path_api:
            return self.__indexFromPath(*args, **kwargs)
        else:
            return self.__index(*args, **kwargs)

    def __index(self, row, column, parent=qt.QModelIndex()):
        return qt.QSortFilterProxyModel.index(self, row, column, parent)

    def __indexFromPath(self, path, column=0):
        """
        Uses the index(str) C++ API

        :rtype: qt.QModelIndex
        """
        if path == "":
            return qt.QModelIndex()

        index = self.sourceModel().index(path, column)
        index = self.mapFromSource(index)
        return index

    def lessThan(self, leftSourceIndex, rightSourceIndex):
        sourceModel = self.sourceModel()
        sortColumn = self.sortColumn()
        if sortColumn == _RawFileSystemModel.NAME_COLUMN:
            leftItem = sourceModel._item(leftSourceIndex)
            rightItem = sourceModel._item(rightSourceIndex)
            if sys.platform != "darwin":
                # Sort directories before files
                leftIsDir = leftItem.isDir()
                rightIsDir = rightItem.isDir()
                if leftIsDir ^ rightIsDir:
                    return leftIsDir
            return leftItem.fileName().lower() < rightItem.fileName().lower()
        elif sortColumn == _RawFileSystemModel.SIZE_COLUMN:
            left = sourceModel.fileInfo(leftSourceIndex)
            right = sourceModel.fileInfo(rightSourceIndex)
            return left.size() < right.size()
        elif sortColumn == _RawFileSystemModel.TYPE_COLUMN:
            left = sourceModel.type(leftSourceIndex)
            right = sourceModel.type(rightSourceIndex)
            return left < right
        elif sortColumn == _RawFileSystemModel.LAST_MODIFIED_COLUMN:
            left = sourceModel.fileInfo(leftSourceIndex)
            right = sourceModel.fileInfo(rightSourceIndex)
            return left.lastModified() < right.lastModified()
        else:
            _logger.warning("Unsupported sorted column %d", sortColumn)

        return False

    def __filtersAccepted(self, item, filters):
        """
        Check individual flag filters.
        """
        if not (filters & (qt.QDir.Dirs | qt.QDir.AllDirs)):
            # Hide dirs
            if item.isDir():
                return False
        if not (filters & qt.QDir.Files):
            # Hide files
            if item.isFile():
                return False
        if not (filters & qt.QDir.Drives):
            # Hide drives
            if item.isDrive():
                return False

        fileInfo = item.fileInfo()
        if fileInfo is None:
            return False

        filterPermissions = (filters & qt.QDir.PermissionMask) != 0
        if filterPermissions and (filters & (qt.QDir.Dirs | qt.QDir.Files)):
            if (filters & qt.QDir.Readable):
                # Hide unreadable
                if not fileInfo.isReadable():
                    return False
            if (filters & qt.QDir.Writable):
                # Hide unwritable
                if not fileInfo.isWritable():
                    return False
            if (filters & qt.QDir.Executable):
                # Hide unexecutable
                if not fileInfo.isExecutable():
                    return False

        if (filters & qt.QDir.NoSymLinks):
            # Hide sym links
            if fileInfo.isSymLink():
                return False

        if not (filters & qt.QDir.System):
            # Hide system
            if not item.isDir() and not item.isFile():
                return False

        fileName = item.fileName()
        isDot = fileName == "."
        isDotDot = fileName == ".."

        if not (filters & qt.QDir.Hidden):
            # Hide hidden
            if not (isDot or isDotDot) and fileInfo.isHidden():
                return False

        if filters & (qt.QDir.NoDot | qt.QDir.NoDotDot | qt.QDir.NoDotAndDotDot):
            # Hide parent/self references
            if filters & qt.QDir.NoDot:
                if isDot:
                    return False
            if filters & qt.QDir.NoDotDot:
                if isDotDot:
                    return False
            if filters & qt.QDir.NoDotAndDotDot:
                if isDot or isDotDot:
                    return False

        return True

    def filterAcceptsRow(self, sourceRow, sourceParent):
        if not sourceParent.isValid():
            return True

        sourceModel = self.sourceModel()
        index = sourceModel.index(sourceRow, 0, sourceParent)
        if not index.isValid():
            return True
        item = sourceModel._item(index)

        filters = self.__filters

        if item.isDrive():
            # Let say a user always have access to a drive
            # It avoid to access to fileInfo then avoid to freeze the file
            # system
            return True

        if not self.__filtersAccepted(item, filters):
            return False

        if self.__nameFilterDisables:
            return True

        if item.isDir() and (filters & qt.QDir.AllDirs):
            # dont apply the filters to directory names
            return True

        return self.__nameFiltersAccepted(item)

    def __nameFiltersAccepted(self, item):
        if len(self.__nameFilters) == 0:
            return True

        fileName = item.fileName()
        for reg in self.__nameFilters:
            if reg.exactMatch(fileName):
                return True
        return False

    def setNameFilterDisables(self, enable):
        self.__nameFilterDisables = enable
        self.invalidate()

    def nameFilterDisables(self):
        return self.__nameFilterDisables

    def myComputer(self, role=qt.Qt.DisplayRole):
        return self.sourceModel().myComputer(role)

    def setNameFilters(self, filters):
        self.__nameFilters = []
        isCaseSensitive = self.__filters & qt.QDir.CaseSensitive
        caseSensitive = qt.Qt.CaseSensitive if isCaseSensitive else qt.Qt.CaseInsensitive
        for f in filters:
            reg = qt.QRegExp(f, caseSensitive, qt.QRegExp.Wildcard)
            self.__nameFilters.append(reg)
        self.invalidate()

    def nameFilters(self):
        return [f.pattern() for f in self.__nameFilters]

    def filter(self):
        return self.__filters

    def setFilter(self, filters):
        self.__filters = filters
        # In case of change of case sensitivity
        self.setNameFilters(self.nameFilters())
        self.invalidate()

    def setReadOnly(self, enable):
        assert(enable is True)

    def isReadOnly(self):
        return False

    def rootPath(self):
        return self.sourceModel().rootPath()

    def setRootPath(self, path):
        index = self.sourceModel().setRootPath(path)
        index = self.mapFromSource(index)
        return index

    def flags(self, index):
        sourceModel = self.sourceModel()
        index = self.mapToSource(index)
        filters = sourceModel.flags(index)

        if self.__nameFilterDisables and not sourceModel.isDir(index):
            item = sourceModel._item(index)
            if not self.__nameFiltersAccepted(item):
                filters &= ~qt.Qt.ItemIsEnabled

        return filters

    def fileIcon(self, index):
        sourceModel = self.sourceModel()
        index = self.mapToSource(index)
        return sourceModel.fileIcon(index)

    def fileInfo(self, index):
        sourceModel = self.sourceModel()
        index = self.mapToSource(index)
        return sourceModel.fileInfo(index)

    def fileName(self, index):
        sourceModel = self.sourceModel()
        index = self.mapToSource(index)
        return sourceModel.fileName(index)

    def filePath(self, index):
        sourceModel = self.sourceModel()
        index = self.mapToSource(index)
        return sourceModel.filePath(index)

    def isDir(self, index):
        sourceModel = self.sourceModel()
        index = self.mapToSource(index)
        return sourceModel.isDir(index)

    def lastModified(self, index):
        sourceModel = self.sourceModel()
        index = self.mapToSource(index)
        return sourceModel.lastModified(index)

    def permissions(self, index):
        sourceModel = self.sourceModel()
        index = self.mapToSource(index)
        return sourceModel.permissions(index)

    def size(self, index):
        sourceModel = self.sourceModel()
        index = self.mapToSource(index)
        return sourceModel.size(index)

    def type(self, index):
        sourceModel = self.sourceModel()
        index = self.mapToSource(index)
        return sourceModel.type(index)



# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "06/07/2018"

from typing import Optional

from .. import qt
from .Hdf5Node import Hdf5Node
import silx.io.utils


class Hdf5LoadingItem(Hdf5Node):
    """Item displayed when an Hdf5Node is loading.

    At the end of the loading this item is replaced by the loaded one.
    """

    def __init__(
        self,
        text,
        parent,
        animatedIcon,
        openedPath: Optional[str] = None,
    ):
        """Constructor"""
        Hdf5Node.__init__(self, parent, openedPath=openedPath)
        self.__text = text
        self.__animatedIcon = animatedIcon
        self.__animatedIcon.register(self)

    @property
    def obj(self):
        return None

    @property
    def h5Class(self):
        """Returns the class of the stored object.

        :rtype: silx.io.utils.H5Type
        """
        return silx.io.utils.H5Type.FILE

    def dataName(self, role):
        if role == qt.Qt.DecorationRole:
            return self.__animatedIcon.currentIcon()
        if role == qt.Qt.TextAlignmentRole:
            return qt.Qt.AlignTop | qt.Qt.AlignLeft
        if role == qt.Qt.DisplayRole:
            return self.__text
        return None

    def dataDescription(self, role):
        if role == qt.Qt.DecorationRole:
            return None
        if role == qt.Qt.TextAlignmentRole:
            return qt.Qt.AlignTop | qt.Qt.AlignLeft
        if role == qt.Qt.DisplayRole:
            return "Loading..."
        return None



# /*##########################################################################
#
# Copyright (c) 2018-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides functions to access Qt C++ object state:

- :func:`isValid` to check whether a QObject C++ pointer is valid.
- :func:`createdByPython` to check if a QObject was created from Python.
- :func:`ownedByPython` to check if a QObject is currently owned by Python.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "08/10/2018"


from . import _qt as qt


if qt.BINDING == 'PyQt5':
    try:
        from PyQt5.sip import isdeleted as _isdeleted  # noqa
        from PyQt5.sip import ispycreated as createdByPython  # noqa
        from PyQt5.sip import ispyowned as ownedByPython  # noqa
    except ImportError:
        from sip import isdeleted as _isdeleted  # noqa
        from sip import ispycreated as createdByPython  # noqa
        from sip import ispyowned as ownedByPython  # noqa


    def isValid(obj):
        """Returns True if underlying C++ object is valid.

        :param QObject obj:
        :rtype: bool
        """
        return not _isdeleted(obj)

elif qt.BINDING == 'PySide2':
    try:
        from PySide2.shiboken2 import isValid  # noqa
        from PySide2.shiboken2 import createdByPython  # noqa
        from PySide2.shiboken2 import ownedByPython  # noqa
    except ImportError:
        from shiboken2 import isValid  # noqa
        from shiboken2 import createdByPython  # noqa
        from shiboken2 import ownedByPython  # noqa

elif qt.BINDING == 'PySide6':
    from shiboken6 import isValid, createdByPython, ownedByPython  # noqa

elif qt.BINDING == 'PyQt6':
    from PyQt6.sip import isdeleted as _isdeleted  # noqa
    from PyQt6.sip import ispycreated as createdByPython  # noqa
    from PyQt6.sip import ispyowned as ownedByPython  # noqa

    def isValid(obj):
        """Returns True if underlying C++ object is valid.

        :param QObject obj:
        :rtype: bool
        """
        return not _isdeleted(obj)



else:
    raise ImportError("Unsupported Qt binding %s" % qt.BINDING)

__all__ = ['isValid', 'createdByPython', 'ownedByPython']



# /*##########################################################################
#
# Copyright (c) 2017-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This package provides a class sharred by widget from the
data module to format data as text in the same way."""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "24/07/2018"

import logging
import numbers

import numpy

from silx.gui import qt

import h5py


_logger = logging.getLogger(__name__)


class TextFormatter(qt.QObject):
    """Formatter to convert data to string.

    The method :meth:`toString` returns a formatted string from an input data
    using parameters set to this object.

    It support most python and numpy data, expecting dictionary. Unsupported
    data are displayed using the string representation of the object (`str`).

    It provides a set of parameters to custom the formatting of integer and
    float values (:meth:`setIntegerFormat`, :meth:`setFloatFormat`).

    It also allows to custom the use of quotes to display text data
    (:meth:`setUseQuoteForText`), and custom unit used to display imaginary
    numbers (:meth:`setImaginaryUnit`).

    The object emit an event `formatChanged` every time a parametter is
    changed.
    """

    formatChanged = qt.Signal()
    """Emitted when properties of the formatter change."""

    def __init__(self, parent=None, formatter=None):
        """
        Constructor

        :param qt.QObject parent: Owner of the object
        :param TextFormatter formatter: Instantiate this object from the
            formatter
        """
        qt.QObject.__init__(self, parent)
        if formatter is not None:
            self.__integerFormat = formatter.integerFormat()
            self.__floatFormat = formatter.floatFormat()
            self.__useQuoteForText = formatter.useQuoteForText()
            self.__imaginaryUnit = formatter.imaginaryUnit()
            self.__enumFormat = formatter.enumFormat()
        else:
            self.__integerFormat = "%d"
            self.__floatFormat = "%g"
            self.__useQuoteForText = True
            self.__imaginaryUnit = u"j"
            self.__enumFormat = u"%(name)s(%(value)d)"

    def integerFormat(self):
        """Returns the format string controlling how the integer data
        are formated by this object.

        This is the C-style format string used by python when formatting
        strings with the modulus operator.

        :rtype: str
        """
        return self.__integerFormat

    def setIntegerFormat(self, value):
        """Set format string controlling how the integer data are
        formated by this object.

        :param str value: Format string (e.g. "%d", "%i", "%08i").
            This is the C-style format string used by python when formatting
            strings with the modulus operator.
        """
        if self.__integerFormat == value:
            return
        self.__integerFormat = value
        self.formatChanged.emit()

    def floatFormat(self):
        """Returns the format string controlling how the floating-point data
        are formated by this object.

        This is the C-style format string used by python when formatting
        strings with the modulus operator.

        :rtype: str
        """
        return self.__floatFormat

    def setFloatFormat(self, value):
        """Set format string controlling how the floating-point data are
        formated by this object.

        :param str value: Format string (e.g. "%.3f", "%d", "%-10.2f",
            "%10.3e").
            This is the C-style format string used by python when formatting
            strings with the modulus operator.
        """
        if self.__floatFormat == value:
            return
        self.__floatFormat = value
        self.formatChanged.emit()

    def useQuoteForText(self):
        """Returns true if the string data are formatted using double quotes.

        Else, no quotes are used.
        """
        return self.__integerFormat

    def setUseQuoteForText(self, useQuote):
        """Set the use of quotes to delimit string data.

        :param bool useQuote: True to use quotes.
        """
        if self.__useQuoteForText == useQuote:
            return
        self.__useQuoteForText = useQuote
        self.formatChanged.emit()

    def imaginaryUnit(self):
        """Returns the unit display for imaginary numbers.

        :rtype: str
        """
        return self.__imaginaryUnit

    def setImaginaryUnit(self, imaginaryUnit):
        """Set the unit display for imaginary numbers.

        :param str imaginaryUnit: Unit displayed after imaginary numbers
        """
        if self.__imaginaryUnit == imaginaryUnit:
            return
        self.__imaginaryUnit = imaginaryUnit
        self.formatChanged.emit()

    def setEnumFormat(self, value):
        """Set format string controlling how the enum data are
        formated by this object.

        :param str value: Format string (e.g. "%(name)s(%(value)d)").
            This is the C-style format string used by python when formatting
            strings with the modulus operator.
        """
        if self.__enumFormat == value:
            return
        self.__enumFormat = value
        self.formatChanged.emit()

    def enumFormat(self):
        """Returns the format string controlling how the enum data
        are formated by this object.

        This is the C-style format string used by python when formatting
        strings with the modulus operator.

        :rtype: str
        """
        return self.__enumFormat

    def __formatText(self, text):
        if self.__useQuoteForText:
            text = "\"%s\"" % text.replace("\\", "\\\\").replace("\"", "\\\"")
        return text

    def __formatBinary(self, data):
        if isinstance(data, numpy.void):
            data = data.item()
            if isinstance(data, numpy.ndarray):
                # Before numpy 1.15.0 the item API was returning a numpy array
                data = data.astype(numpy.uint8)
            else:
                # Now it is supposed to be a bytes type
                pass
        data = ["\\x%02X" % d for d in data]
        if self.__useQuoteForText:
            return "b\"%s\"" % "".join(data)
        else:
            return "".join(data)

    def __formatSafeAscii(self, data):
        data = [chr(d) if (d > 0x20 and d < 0x7F) else "\\x%02X" % d for d in data]
        if self.__useQuoteForText:
            data = [c if c != '"' else "\\" + c for c in data]
            return "b\"%s\"" % "".join(data)
        else:
            return "".join(data)

    def __formatCharString(self, data):
        """Format text of char.

        From the specifications we expect to have ASCII, but we also allow
        CP1252 in some ceases as fallback.

        If no encoding fits, it will display a readable ASCII chars, with
        escaped chars (using the python syntax) for non decoded characters.

        :param data: A binary string of char expected in ASCII
        :rtype: str
        """
        try:
            text = "%s" % data.decode("ascii")
            return self.__formatText(text)
        except UnicodeDecodeError:
            # Here we can spam errors, this is definitly a badly
            # generated file
            _logger.error("Invalid ASCII string %s.", data)
            if data == b"\xB0":
                _logger.error("Fallback using cp1252 encoding")
                return self.__formatText(u"\u00B0")
        return self.__formatSafeAscii(data)

    def __formatH5pyObject(self, data, dtype):
        # That's an HDF5 object
        ref = h5py.check_dtype(ref=dtype)
        if ref is not None:
            if bool(data):
                return "REF"
            else:
                return "NULL_REF"
        vlen = h5py.check_dtype(vlen=dtype)
        if vlen is not None:
            if vlen == str:
                # HDF5 UTF8
                # With h5py>=3 reading dataset returns bytes
                if isinstance(data, (bytes, numpy.bytes_)):
                    try:
                        data = data.decode("utf-8")
                    except UnicodeDecodeError:
                        self.__formatSafeAscii(data)
                return self.__formatText(data)
            elif vlen == bytes:
                # HDF5 ASCII
                return self.__formatCharString(data)
            elif isinstance(vlen, numpy.dtype):
                return self.toString(data, vlen)
        return None

    def toString(self, data, dtype=None):
        """Format a data into a string using formatter options

        :param object data: Data to render
        :param dtype: enforce a dtype (mostly used to remember the h5py dtype,
             special h5py dtypes are not propagated from array to items)
        :rtype: str
        """
        if isinstance(data, tuple):
            text = [self.toString(d) for d in data]
            return "(" + " ".join(text) + ")"
        elif isinstance(data, list):
            text = [self.toString(d) for d in data]
            return "[" + " ".join(text) + "]"
        elif isinstance(data, numpy.ndarray):
            if dtype is None:
                dtype = data.dtype
            if data.shape == ():
                # it is a scaler
                return self.toString(data[()], dtype)
            else:
                text = [self.toString(d, dtype) for d in data]
                return "[" + " ".join(text) + "]"
        if dtype is not None and dtype.kind == 'O':
            text = self.__formatH5pyObject(data, dtype)
            if text is not None:
                return text
        elif isinstance(data, numpy.void):
            if dtype is None:
                dtype = data.dtype
            if dtype.fields is not None:
                text = []
                for index, field in enumerate(dtype.fields.items()):
                    text.append(field[0] + ":" + self.toString(data[index], field[1][0]))
                return "(" + " ".join(text) + ")"
            return self.__formatBinary(data)
        elif isinstance(data, (numpy.unicode_, str)):
            return self.__formatText(data)
        elif isinstance(data, (numpy.string_, bytes)):
            if dtype is None and hasattr(data, "dtype"):
                dtype = data.dtype
            if dtype is not None:
                # Maybe a sub item from HDF5
                if dtype.kind == 'S':
                    return self.__formatCharString(data)
                elif dtype.kind == 'O':
                    text = self.__formatH5pyObject(data, dtype)
                    if text is not None:
                        return text
            try:
                # Try ascii/utf-8
                text = "%s" % data.decode("utf-8")
                return self.__formatText(text)
            except UnicodeDecodeError:
                pass
            return self.__formatBinary(data)
        elif isinstance(data, str):
            text = "%s" % data
            return self.__formatText(text)
        elif isinstance(data, (numpy.integer)):
            if dtype is None:
                dtype = data.dtype
            enumType = h5py.check_dtype(enum=dtype)
            if enumType is not None:
                for key, value in enumType.items():
                    if value == data:
                        result = {}
                        result["name"] = key
                        result["value"] = data
                        return self.__enumFormat % result
            return self.__integerFormat % data
        elif isinstance(data, (numbers.Integral)):
            return self.__integerFormat % data
        elif isinstance(data, (numbers.Real, numpy.floating)):
            # It have to be done before complex checking
            return self.__floatFormat % data
        elif isinstance(data, (numpy.complexfloating, numbers.Complex)):
            text = ""
            if data.real != 0:
                text += self.__floatFormat % data.real
            if data.real != 0 and data.imag != 0:
                if data.imag < 0:
                    template = self.__floatFormat + " - " + self.__floatFormat + self.__imaginaryUnit
                    params = (data.real, -data.imag)
                else:
                    template = self.__floatFormat + " + " + self.__floatFormat + self.__imaginaryUnit
                    params = (data.real, data.imag)
            else:
                if data.imag != 0:
                    template = self.__floatFormat + self.__imaginaryUnit
                    params = (data.imag)
                else:
                    template = self.__floatFormat
                    params = (data.real)
            return template % params
        elif isinstance(data, h5py.h5r.Reference):
            dtype = h5py.special_dtype(ref=h5py.Reference)
            text = self.__formatH5pyObject(data, dtype)
            return text
        elif isinstance(data, h5py.h5r.RegionReference):
            dtype = h5py.special_dtype(ref=h5py.RegionReference)
            text = self.__formatH5pyObject(data, dtype)
            return text
        elif isinstance(data, numpy.object_) or dtype is not None:
            if dtype is None:
                dtype = data.dtype
            text = self.__formatH5pyObject(data, dtype)
            if text is not None:
                return text
            # That's a numpy object
            return str(data)
        return str(data)



# /*##########################################################################
#
# Copyright (c) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Tests for debug module"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "27/02/2018"


import unittest
from silx.utils import debug
from silx.utils import testutils


@debug.log_all_methods
class _Foobar(object):

    def a(self):
        return None

    def b(self):
        return self.a()

    def random_args(self, *args, **kwargs):
        return args, kwargs

    def named_args(self, a, b):
        return a + 1, b + 1


class TestDebug(unittest.TestCase):
    """Tests for debug module."""

    def logB(self):
        """
        Can be used to check the log output using:
        `./run_tests.py silx.utils.test.test_debug.TestDebug.logB -v`
        """
        print()
        test = _Foobar()
        test.b()

    @testutils.validate_logging(debug.debug_logger.name, warning=2)
    def testMethod(self):
        test = _Foobar()
        test.a()

    @testutils.validate_logging(debug.debug_logger.name, warning=4)
    def testInterleavedMethod(self):
        test = _Foobar()
        test.b()

    @testutils.validate_logging(debug.debug_logger.name, warning=2)
    def testNamedArgument(self):
        # Arguments arre still provided to the patched method
        test = _Foobar()
        result = test.named_args(10, 11)
        self.assertEqual(result, (11, 12))

    @testutils.validate_logging(debug.debug_logger.name, warning=2)
    def testRandomArguments(self):
        # Arguments arre still provided to the patched method
        test = _Foobar()
        result = test.random_args("foo", 50, a=10, b=100)
        self.assertEqual(result[0], ("foo", 50))
        self.assertEqual(result[1], {"a": 10, "b": 100})



# /*##########################################################################
#
# Copyright (c) 2004-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
The class :class:`.PlotAction` help the creation of a qt.QAction associated
with a :class:`.PlotWidget`.
"""

__authors__ = ["V.A. Sole", "T. Vincent", "P. Knobel"]
__license__ = "MIT"
__date__ = "03/01/2018"


import weakref
from silx.gui import icons
from silx.gui import qt


class PlotAction(qt.QAction):
    """Base class for QAction that operates on a PlotWidget.

    :param plot: :class:`.PlotWidget` instance on which to operate.
    :param icon: QIcon or str name of icon to use
    :param str text: The name of this action to be used for menu label
    :param str tooltip: The text of the tooltip
    :param triggered: The callback to connect to the action's triggered
                      signal or None for no callback.
    :param bool checkable: True for checkable action, False otherwise (default)
    :param parent: See :class:`QAction`.
    """

    def __init__(self, plot, icon, text, tooltip=None,
                 triggered=None, checkable=False, parent=None):
        assert plot is not None
        self._plotRef = weakref.ref(plot)

        if not isinstance(icon, qt.QIcon):
            # Try with icon as a string and load corresponding icon
            icon = icons.getQIcon(icon)

        super(PlotAction, self).__init__(icon, text, parent)

        if tooltip is not None:
            self.setToolTip(tooltip)

        self.setCheckable(checkable)

        if triggered is not None:
            self.triggered[bool].connect(triggered)

    @property
    def plot(self):
        """The :class:`.PlotWidget` this action group is controlling."""
        return self._plotRef()



# /*##########################################################################
# Copyright (C) 2016-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""Tests for NXdata parsing"""

__authors__ = ["P. Knobel"]
__license__ = "MIT"
__date__ = "24/03/2020"


import tempfile
import unittest

import h5py
import numpy
import pytest

from .. import nxdata
from ..dictdump import dicttoh5


text_dtype = h5py.special_dtype(vlen=str)


class TestNXdata(unittest.TestCase):
    def setUp(self):
        tmp = tempfile.NamedTemporaryFile(prefix="nxdata_examples_", suffix=".h5", delete=True)
        tmp.file.close()
        self.h5fname = tmp.name
        self.h5f = h5py.File(tmp.name, "w")

        # SCALARS
        g0d = self.h5f.create_group("scalars")

        g0d0 = g0d.create_group("0D_scalar")
        g0d0.attrs["NX_class"] = "NXdata"
        g0d0.attrs["signal"] = "scalar"
        g0d0.create_dataset("scalar", data=10)
        g0d0.create_dataset("scalar_errors", data=0.1)

        g0d1 = g0d.create_group("2D_scalars")
        g0d1.attrs["NX_class"] = "NXdata"
        g0d1.attrs["signal"] = "scalars"
        ds = g0d1.create_dataset("scalars", data=numpy.arange(3 * 10).reshape((3, 10)))
        ds.attrs["interpretation"] = "scalar"

        g0d1 = g0d.create_group("4D_scalars")
        g0d1.attrs["NX_class"] = "NXdata"
        g0d1.attrs["signal"] = "scalars"
        ds = g0d1.create_dataset("scalars", data=numpy.arange(2 * 2 * 3 * 10).reshape((2, 2, 3, 10)))
        ds.attrs["interpretation"] = "scalar"

        # SPECTRA
        g1d = self.h5f.create_group("spectra")

        g1d0 = g1d.create_group("1D_spectrum")
        g1d0.attrs["NX_class"] = "NXdata"
        g1d0.attrs["signal"] = "count"
        g1d0.attrs["auxiliary_signals"] = numpy.array(["count2", "count3"],
                                                      dtype=text_dtype)
        g1d0.attrs["axes"] = "energy_calib"
        g1d0.attrs["uncertainties"] = numpy.array(["energy_errors", ],
                                                  dtype=text_dtype)
        g1d0.create_dataset("count", data=numpy.arange(10))
        g1d0.create_dataset("count2", data=0.5 * numpy.arange(10))
        d = g1d0.create_dataset("count3", data=0.4 * numpy.arange(10))
        d.attrs["long_name"] = "3rd counter"
        g1d0.create_dataset("title", data="Title as dataset (like nexpy)")
        g1d0.create_dataset("energy_calib", data=(10, 5))  # 10 * idx + 5
        g1d0.create_dataset("energy_errors", data=3.14 * numpy.random.rand(10))

        g1d1 = g1d.create_group("2D_spectra")
        g1d1.attrs["NX_class"] = "NXdata"
        g1d1.attrs["signal"] = "counts"
        ds = g1d1.create_dataset("counts", data=numpy.arange(3 * 10).reshape((3, 10)))
        ds.attrs["interpretation"] = "spectrum"

        g1d2 = g1d.create_group("4D_spectra")
        g1d2.attrs["NX_class"] = "NXdata"
        g1d2.attrs["signal"] = "counts"
        g1d2.attrs["axes"] = numpy.array(["energy", ], dtype=text_dtype)
        ds = g1d2.create_dataset("counts", data=numpy.arange(2 * 2 * 3 * 10).reshape((2, 2, 3, 10)))
        ds.attrs["interpretation"] = "spectrum"
        ds = g1d2.create_dataset("errors", data=4.5 * numpy.random.rand(2, 2, 3, 10))
        ds = g1d2.create_dataset("energy", data=5 + 10 * numpy.arange(15),
                                 shuffle=True, compression="gzip")
        ds.attrs["long_name"] = "Calibrated energy"
        ds.attrs["first_good"] = 3
        ds.attrs["last_good"] = 12
        g1d2.create_dataset("energy_errors", data=10 * numpy.random.rand(15))

        # IMAGES
        g2d = self.h5f.create_group("images")

        g2d0 = g2d.create_group("2D_regular_image")
        g2d0.attrs["NX_class"] = "NXdata"
        g2d0.attrs["signal"] = "image"
        g2d0.attrs["auxiliary_signals"] = "image2"
        g2d0.attrs["axes"] = numpy.array(["rows_calib", "columns_coordinates"],
                                         dtype=text_dtype)
        g2d0.create_dataset("image", data=numpy.arange(4 * 6).reshape((4, 6)))
        g2d0.create_dataset("image2", data=numpy.arange(4 * 6).reshape((4, 6)))
        ds = g2d0.create_dataset("rows_calib", data=(10, 5))
        ds.attrs["long_name"] = "Calibrated Y"
        g2d0.create_dataset("columns_coordinates", data=0.5 + 0.02 * numpy.arange(6))

        g2d1 = g2d.create_group("2D_irregular_data")
        g2d1.attrs["NX_class"] = "NXdata"
        g2d1.attrs["signal"] = "data"
        g2d1.attrs["title"] = "Title as group attr"
        g2d1.attrs["axes"] = numpy.array(["rows_coordinates", "columns_coordinates"],
                                         dtype=text_dtype)
        g2d1.create_dataset("data", data=numpy.arange(64 * 128).reshape((64, 128)))
        g2d1.create_dataset("rows_coordinates", data=numpy.arange(64) + numpy.random.rand(64))
        g2d1.create_dataset("columns_coordinates", data=numpy.arange(128) + 2.5 * numpy.random.rand(128))

        g2d2 = g2d.create_group("3D_images")
        g2d2.attrs["NX_class"] = "NXdata"
        g2d2.attrs["signal"] = "images"
        ds = g2d2.create_dataset("images", data=numpy.arange(2 * 4 * 6).reshape((2, 4, 6)))
        ds.attrs["interpretation"] = "image"

        g2d3 = g2d.create_group("5D_images")
        g2d3.attrs["NX_class"] = "NXdata"
        g2d3.attrs["signal"] = "images"
        g2d3.attrs["axes"] = numpy.array(["rows_coordinates", "columns_coordinates"],
                                         dtype=text_dtype)
        ds = g2d3.create_dataset("images", data=numpy.arange(2 * 2 * 2 * 4 * 6).reshape((2, 2, 2, 4, 6)))
        ds.attrs["interpretation"] = "image"
        g2d3.create_dataset("rows_coordinates", data=5 + 10 * numpy.arange(4))
        g2d3.create_dataset("columns_coordinates", data=0.5 + 0.02 * numpy.arange(6))

        g2d4 = g2d.create_group("RGBA_image")
        g2d4.attrs["NX_class"] = "NXdata"
        g2d4.attrs["signal"] = "image"
        g2d4.attrs["axes"] = numpy.array(["rows_calib", "columns_coordinates"],
                                         dtype=text_dtype)
        rgba_image = numpy.linspace(0, 1, num=7*8*3).reshape((7, 8, 3))
        rgba_image[:, :, 1] = 1 - rgba_image[:, :, 1]      # invert G channel to add some color
        ds = g2d4.create_dataset("image", data=rgba_image)
        ds.attrs["interpretation"] = "rgba-image"
        ds = g2d4.create_dataset("rows_calib", data=(10, 5))
        ds.attrs["long_name"] = "Calibrated Y"
        g2d4.create_dataset("columns_coordinates", data=0.5+0.02*numpy.arange(8))

        # SCATTER
        g = self.h5f.create_group("scatters")

        gd0 = g.create_group("x_y_scatter")
        gd0.attrs["NX_class"] = "NXdata"
        gd0.attrs["signal"] = "y"
        gd0.attrs["axes"] = numpy.array(["x", ], dtype=text_dtype)
        gd0.create_dataset("y", data=numpy.random.rand(128) - 0.5)
        gd0.create_dataset("x", data=2 * numpy.random.rand(128))
        gd0.create_dataset("x_errors", data=0.05 * numpy.random.rand(128))
        gd0.create_dataset("errors", data=0.05 * numpy.random.rand(128))

        gd1 = g.create_group("x_y_value_scatter")
        gd1.attrs["NX_class"] = "NXdata"
        gd1.attrs["signal"] = "values"
        gd1.attrs["axes"] = numpy.array(["x", "y"], dtype=text_dtype)
        gd1.create_dataset("values", data=3.14 * numpy.random.rand(128))
        gd1.create_dataset("y", data=numpy.random.rand(128))
        gd1.create_dataset("y_errors", data=0.02 * numpy.random.rand(128))
        gd1.create_dataset("x", data=numpy.random.rand(128))
        gd1.create_dataset("x_errors", data=0.02 * numpy.random.rand(128))

    def tearDown(self):
        self.h5f.close()

    def testValidity(self):
        for group in self.h5f:
            for subgroup in self.h5f[group]:
                self.assertTrue(
                        nxdata.is_valid_nxdata(self.h5f[group][subgroup]),
                        "%s/%s not found to be a valid NXdata group" % (group, subgroup))

    def testScalars(self):
        nxd = nxdata.NXdata(self.h5f["scalars/0D_scalar"])
        self.assertTrue(nxd.signal_is_0d)
        self.assertEqual(nxd.signal[()], 10)
        self.assertEqual(nxd.axes_names, [])
        self.assertEqual(nxd.axes_dataset_names, [])
        self.assertEqual(nxd.axes, [])
        self.assertIsNotNone(nxd.errors)
        self.assertFalse(nxd.is_scatter or nxd.is_x_y_value_scatter)
        self.assertIsNone(nxd.interpretation)

        nxd = nxdata.NXdata(self.h5f["scalars/2D_scalars"])
        self.assertTrue(nxd.signal_is_2d)
        self.assertEqual(nxd.signal[1, 2], 12)
        self.assertEqual(nxd.axes_names, [None, None])
        self.assertEqual(nxd.axes_dataset_names, [None, None])
        self.assertEqual(nxd.axes, [None, None])
        self.assertIsNone(nxd.errors)
        self.assertFalse(nxd.is_scatter or nxd.is_x_y_value_scatter)
        self.assertEqual(nxd.interpretation, "scalar")

        nxd = nxdata.NXdata(self.h5f["scalars/4D_scalars"])
        self.assertFalse(nxd.signal_is_0d or nxd.signal_is_1d or
                         nxd.signal_is_2d or nxd.signal_is_3d)
        self.assertEqual(nxd.signal[1, 0, 1, 4], 74)
        self.assertEqual(nxd.axes_names, [None, None, None, None])
        self.assertEqual(nxd.axes_dataset_names, [None, None, None, None])
        self.assertEqual(nxd.axes, [None, None, None, None])
        self.assertIsNone(nxd.errors)
        self.assertFalse(nxd.is_scatter or nxd.is_x_y_value_scatter)
        self.assertEqual(nxd.interpretation, "scalar")

    def testSpectra(self):
        nxd = nxdata.NXdata(self.h5f["spectra/1D_spectrum"])
        self.assertTrue(nxd.signal_is_1d)
        self.assertTrue(nxd.is_curve)
        self.assertTrue(numpy.array_equal(numpy.array(nxd.signal),
                                          numpy.arange(10)))
        self.assertEqual(nxd.axes_names, ["energy_calib"])
        self.assertEqual(nxd.axes_dataset_names, ["energy_calib"])
        self.assertEqual(nxd.axes[0][0], 10)
        self.assertEqual(nxd.axes[0][1], 5)
        self.assertIsNone(nxd.errors)
        self.assertFalse(nxd.is_scatter or nxd.is_x_y_value_scatter)
        self.assertIsNone(nxd.interpretation)
        self.assertEqual(nxd.title, "Title as dataset (like nexpy)")

        self.assertEqual(nxd.auxiliary_signals_dataset_names,
                         ["count2", "count3"])
        self.assertEqual(nxd.auxiliary_signals_names,
                         ["count2", "3rd counter"])
        self.assertAlmostEqual(nxd.auxiliary_signals[1][2],
                               0.8)      # numpy.arange(10) * 0.4

        nxd = nxdata.NXdata(self.h5f["spectra/2D_spectra"])
        self.assertTrue(nxd.signal_is_2d)
        self.assertTrue(nxd.is_curve)
        self.assertEqual(nxd.axes_names, [None, None])
        self.assertEqual(nxd.axes_dataset_names, [None, None])
        self.assertEqual(nxd.axes, [None, None])
        self.assertIsNone(nxd.errors)
        self.assertFalse(nxd.is_scatter or nxd.is_x_y_value_scatter)
        self.assertEqual(nxd.interpretation, "spectrum")

        nxd = nxdata.NXdata(self.h5f["spectra/4D_spectra"])
        self.assertFalse(nxd.signal_is_0d or nxd.signal_is_1d or
                         nxd.signal_is_2d or nxd.signal_is_3d)
        self.assertTrue(nxd.is_curve)
        self.assertEqual(nxd.axes_names,
                         [None, None, None, "Calibrated energy"])
        self.assertEqual(nxd.axes_dataset_names,
                         [None, None, None, "energy"])
        self.assertEqual(nxd.axes[:3], [None, None, None])
        self.assertEqual(nxd.axes[3].shape, (10, ))    # dataset shape (15, ) sliced [3:12]
        self.assertIsNotNone(nxd.errors)
        self.assertEqual(nxd.errors.shape, (2, 2, 3, 10))
        self.assertFalse(nxd.is_scatter or nxd.is_x_y_value_scatter)
        self.assertEqual(nxd.interpretation, "spectrum")
        self.assertEqual(nxd.get_axis_errors("energy").shape,
                         (10,))
        # test getting axis errors by long_name
        self.assertTrue(numpy.array_equal(nxd.get_axis_errors("Calibrated energy"),
                                          nxd.get_axis_errors("energy")))
        self.assertTrue(numpy.array_equal(nxd.get_axis_errors(b"Calibrated energy"),
                                          nxd.get_axis_errors("energy")))

    def testImages(self):
        nxd = nxdata.NXdata(self.h5f["images/2D_regular_image"])
        self.assertTrue(nxd.signal_is_2d)
        self.assertTrue(nxd.is_image)
        self.assertEqual(nxd.axes_names, ["Calibrated Y", "columns_coordinates"])
        self.assertEqual(list(nxd.axes_dataset_names),
                         ["rows_calib", "columns_coordinates"])
        self.assertIsNone(nxd.errors)
        self.assertFalse(nxd.is_scatter or nxd.is_x_y_value_scatter)
        self.assertIsNone(nxd.interpretation)
        self.assertEqual(len(nxd.auxiliary_signals), 1)
        self.assertEqual(nxd.auxiliary_signals_names, ["image2"])

        nxd = nxdata.NXdata(self.h5f["images/2D_irregular_data"])
        self.assertTrue(nxd.signal_is_2d)
        self.assertTrue(nxd.is_image)

        self.assertEqual(nxd.axes_dataset_names, nxd.axes_names)
        self.assertEqual(list(nxd.axes_dataset_names),
                         ["rows_coordinates", "columns_coordinates"])
        self.assertEqual(len(nxd.axes), 2)
        self.assertIsNone(nxd.errors)
        self.assertFalse(nxd.is_scatter or nxd.is_x_y_value_scatter)
        self.assertIsNone(nxd.interpretation)
        self.assertEqual(nxd.title, "Title as group attr")

        nxd = nxdata.NXdata(self.h5f["images/5D_images"])
        self.assertTrue(nxd.is_image)
        self.assertFalse(nxd.signal_is_0d or nxd.signal_is_1d or
                         nxd.signal_is_2d or nxd.signal_is_3d)
        self.assertEqual(nxd.axes_names,
                         [None, None, None, 'rows_coordinates', 'columns_coordinates'])
        self.assertEqual(nxd.axes_dataset_names,
                         [None, None, None, 'rows_coordinates', 'columns_coordinates'])
        self.assertIsNone(nxd.errors)
        self.assertFalse(nxd.is_scatter or nxd.is_x_y_value_scatter)
        self.assertEqual(nxd.interpretation, "image")

        nxd = nxdata.NXdata(self.h5f["images/RGBA_image"])
        self.assertTrue(nxd.is_image)
        self.assertEqual(nxd.interpretation, "rgba-image")
        self.assertTrue(nxd.signal_is_3d)
        self.assertEqual(nxd.axes_names, ["Calibrated Y",
                                          "columns_coordinates",
                                          None])
        self.assertEqual(list(nxd.axes_dataset_names),
                         ["rows_calib", "columns_coordinates", None])

    def testScatters(self):
        nxd = nxdata.NXdata(self.h5f["scatters/x_y_scatter"])
        self.assertTrue(nxd.signal_is_1d)
        self.assertEqual(nxd.axes_names, ["x"])
        self.assertEqual(nxd.axes_dataset_names,
                         ["x"])
        self.assertIsNotNone(nxd.errors)
        self.assertEqual(nxd.get_axis_errors("x").shape,
                         (128, ))
        self.assertTrue(nxd.is_scatter)
        self.assertFalse(nxd.is_x_y_value_scatter)
        self.assertIsNone(nxd.interpretation)

        nxd = nxdata.NXdata(self.h5f["scatters/x_y_value_scatter"])
        self.assertFalse(nxd.signal_is_1d)
        self.assertTrue(nxd.axes_dataset_names,
                        nxd.axes_names)
        self.assertEqual(nxd.axes_dataset_names,
                         ["x", "y"])
        self.assertEqual(nxd.get_axis_errors("x").shape,
                         (128, ))
        self.assertEqual(nxd.get_axis_errors("y").shape,
                         (128, ))
        self.assertEqual(len(nxd.axes), 2)
        self.assertIsNone(nxd.errors)
        self.assertTrue(nxd.is_scatter)
        self.assertTrue(nxd.is_x_y_value_scatter)
        self.assertIsNone(nxd.interpretation)


class TestLegacyNXdata(unittest.TestCase):
    def setUp(self):
        tmp = tempfile.NamedTemporaryFile(prefix="nxdata_legacy_examples_",
                                          suffix=".h5", delete=True)
        tmp.file.close()
        self.h5fname = tmp.name
        self.h5f = h5py.File(tmp.name, "w")

    def tearDown(self):
        self.h5f.close()

    def testSignalAttrOnDataset(self):
        g = self.h5f.create_group("2D")
        g.attrs["NX_class"] = "NXdata"

        ds0 = g.create_dataset("image0",
                               data=numpy.arange(4 * 6).reshape((4, 6)))
        ds0.attrs["signal"] = 1
        ds0.attrs["long_name"] = "My first image"

        ds1 = g.create_dataset("image1",
                               data=numpy.arange(4 * 6).reshape((4, 6)))
        ds1.attrs["signal"] = "2"
        ds1.attrs["long_name"] = "My 2nd image"

        ds2 = g.create_dataset("image2",
                               data=numpy.arange(4 * 6).reshape((4, 6)))
        ds2.attrs["signal"] = 3

        nxd = nxdata.NXdata(self.h5f["2D"])

        self.assertEqual(nxd.signal_dataset_name, "image0")
        self.assertEqual(nxd.signal_name, "My first image")
        self.assertEqual(nxd.signal.shape,
                         (4, 6))

        self.assertEqual(len(nxd.auxiliary_signals), 2)
        self.assertEqual(nxd.auxiliary_signals[1].shape,
                         (4, 6))

        self.assertEqual(nxd.auxiliary_signals_dataset_names,
                         ["image1", "image2"])
        self.assertEqual(nxd.auxiliary_signals_names,
                         ["My 2nd image", "image2"])

    def testAxesOnSignalDataset(self):
        g = self.h5f.create_group("2D")
        g.attrs["NX_class"] = "NXdata"

        ds0 = g.create_dataset("image0",
                               data=numpy.arange(4 * 6).reshape((4, 6)))
        ds0.attrs["signal"] = 1
        ds0.attrs["axes"] = "yaxis:xaxis"

        ds1 = g.create_dataset("yaxis",
                               data=numpy.arange(4))
        ds2 = g.create_dataset("xaxis",
                               data=numpy.arange(6))

        nxd = nxdata.NXdata(self.h5f["2D"])

        self.assertEqual(nxd.axes_dataset_names,
                         ["yaxis", "xaxis"])
        self.assertTrue(numpy.array_equal(nxd.axes[0],
                                          numpy.arange(4)))
        self.assertTrue(numpy.array_equal(nxd.axes[1],
                                          numpy.arange(6)))

    def testAxesOnAxesDatasets(self):
        g = self.h5f.create_group("2D")
        g.attrs["NX_class"] = "NXdata"

        ds0 = g.create_dataset("image0",
                               data=numpy.arange(4 * 6).reshape((4, 6)))
        ds0.attrs["signal"] = 1
        ds1 = g.create_dataset("yaxis",
                               data=numpy.arange(4))
        ds1.attrs["axis"] = 0
        ds2 = g.create_dataset("xaxis",
                               data=numpy.arange(6))
        ds2.attrs["axis"] = "1"

        nxd = nxdata.NXdata(self.h5f["2D"])
        self.assertEqual(nxd.axes_dataset_names,
                         ["yaxis", "xaxis"])
        self.assertTrue(numpy.array_equal(nxd.axes[0],
                                          numpy.arange(4)))
        self.assertTrue(numpy.array_equal(nxd.axes[1],
                                          numpy.arange(6)))

    def testAsciiUndefinedAxesAttrs(self):
        """Some files may not be using utf8 for str attrs"""
        g = self.h5f.create_group("bytes_attrs")
        g.attrs["NX_class"] = b"NXdata"
        g.attrs["signal"] = b"image0"
        g.attrs["axes"] = b"yaxis", b"."

        g.create_dataset("image0",
                         data=numpy.arange(4 * 6).reshape((4, 6)))
        g.create_dataset("yaxis",
                         data=numpy.arange(4))

        nxd = nxdata.NXdata(self.h5f["bytes_attrs"])
        self.assertEqual(nxd.axes_dataset_names,
                         ["yaxis", None])


class TestSaveNXdata(unittest.TestCase):
    def setUp(self):
        tmp = tempfile.NamedTemporaryFile(prefix="nxdata",
                                          suffix=".h5", delete=True)
        tmp.file.close()
        self.h5fname = tmp.name

    def testSimpleSave(self):
        sig = numpy.array([0, 1, 2])
        a0 = numpy.array([2, 3, 4])
        a1 = numpy.array([3, 4, 5])
        nxdata.save_NXdata(filename=self.h5fname,
                           signal=sig,
                           axes=[a0, a1],
                           signal_name="sig",
                           axes_names=["a0", "a1"],
                           nxentry_name="a",
                           nxdata_name="mydata")

        h5f = h5py.File(self.h5fname, "r")
        self.assertTrue(nxdata.is_valid_nxdata(h5f["a/mydata"]))

        nxd = nxdata.NXdata(h5f["/a/mydata"])
        self.assertTrue(numpy.array_equal(nxd.signal,
                                          sig))
        self.assertTrue(numpy.array_equal(nxd.axes[0],
                                          a0))

        h5f.close()

    def testSimplestSave(self):
        sig = numpy.array([0, 1, 2])
        nxdata.save_NXdata(filename=self.h5fname,
                           signal=sig)

        h5f = h5py.File(self.h5fname, "r")

        self.assertTrue(nxdata.is_valid_nxdata(h5f["/entry/data0"]))

        nxd = nxdata.NXdata(h5f["/entry/data0"])
        self.assertTrue(numpy.array_equal(nxd.signal,
                                          sig))
        h5f.close()

    def testSaveDefaultAxesNames(self):
        sig = numpy.array([0, 1, 2])
        a0 = numpy.array([2, 3, 4])
        a1 = numpy.array([3, 4, 5])
        nxdata.save_NXdata(filename=self.h5fname,
                           signal=sig,
                           axes=[a0, a1],
                           signal_name="sig",
                           axes_names=None,
                           axes_long_names=["a", "b"],
                           nxentry_name="a",
                           nxdata_name="mydata")

        h5f = h5py.File(self.h5fname, "r")
        self.assertTrue(nxdata.is_valid_nxdata(h5f["a/mydata"]))

        nxd = nxdata.NXdata(h5f["/a/mydata"])
        self.assertTrue(numpy.array_equal(nxd.signal,
                                          sig))
        self.assertTrue(numpy.array_equal(nxd.axes[0],
                                          a0))
        self.assertEqual(nxd.axes_dataset_names,
                         [u"dim0", u"dim1"])
        self.assertEqual(nxd.axes_names,
                         [u"a", u"b"])

        h5f.close()

    def testSaveToExistingEntry(self):
        h5f = h5py.File(self.h5fname, "w")
        g = h5f.create_group("myentry")
        g.attrs["NX_class"] = "NXentry"
        h5f.close()

        sig = numpy.array([0, 1, 2])
        a0 = numpy.array([2, 3, 4])
        a1 = numpy.array([3, 4, 5])
        nxdata.save_NXdata(filename=self.h5fname,
                           signal=sig,
                           axes=[a0, a1],
                           signal_name="sig",
                           axes_names=["a0", "a1"],
                           nxentry_name="myentry",
                           nxdata_name="toto")

        h5f = h5py.File(self.h5fname, "r")
        self.assertTrue(nxdata.is_valid_nxdata(h5f["myentry/toto"]))

        nxd = nxdata.NXdata(h5f["myentry/toto"])
        self.assertTrue(numpy.array_equal(nxd.signal,
                                          sig))
        self.assertTrue(numpy.array_equal(nxd.axes[0],
                                          a0))
        h5f.close()


class TestGetDefault:
    """Test silx.io.nxdata.get_default function"""

    @pytest.fixture
    def hdf5_file(self, tmp_path):
        with h5py.File(tmp_path / "test_file.h5", "w") as h5f:
            yield h5f

    def testDirectPath(self, hdf5_file):
        dicttoh5(
            {
                ("", "default"): "/nxentry/nxprocess/nxdata",
                "nxentry": {
                    "nxprocess": {
                        "nxdata": {
                            ("", "NX_class"): "NXdata",
                            ("", "signal"): "data",
                            "data": (1, 2, 3),
                        }
                    }
                }
            },
            hdf5_file)
        default = nxdata.get_default(hdf5_file)
        assert isinstance(default, nxdata.NXdata)
        assert default.group.name == "/nxentry/nxprocess/nxdata"

    def testAbsolutePath(self, hdf5_file):
        dicttoh5(
            {
                ("", "default"): "/nxentry",
                "nxentry": {
                    ("", "default"): "/nxentry/nxprocess/nxdata",
                    "nxprocess": {
                        "nxdata": {
                            ("", "NX_class"): "NXdata",
                            ("", "signal"): "data",
                            "data": (1, 2, 3),
                        }
                    }
                }
            },
            hdf5_file)
        default = nxdata.get_default(hdf5_file)
        assert isinstance(default, nxdata.NXdata)
        assert default.group.name == "/nxentry/nxprocess/nxdata"

    def testRelativePath(self, hdf5_file):
        dicttoh5(
            {
                ("", "default"): "nxentry",
                "nxentry": {
                    ("", "default"): "nxdata",
                    "nxdata": {
                        ("", "NX_class"): "NXdata",
                        ("", "signal"): "data",
                        "data": (1, 2, 3),
                    }
                }
            },
            hdf5_file)
        default = nxdata.get_default(hdf5_file)
        assert isinstance(default, nxdata.NXdata)
        assert default.group.name == "/nxentry/nxdata"

    def testRelativePathSubdir(self, hdf5_file):
        dicttoh5(
            {
                ("", "default"): "nxentry",
                "nxentry": {
                    ("", "default"): "nxprocess/nxdata",
                    "nxprocess": {
                        "nxdata": {
                            ("", "NX_class"): "NXdata",
                            ("", "signal"): "data",
                            "data": (1, 2, 3),
                        }
                    }
                }
            },
            hdf5_file)
        default = nxdata.get_default(hdf5_file)
        assert isinstance(default, nxdata.NXdata)
        assert default.group.name == "/nxentry/nxprocess/nxdata"

    def testRecursiveAbsolutePath(self, hdf5_file):
        dicttoh5(
            {
                ("", "default"): "/nxentry",
                "nxentry": {
                    ("", "default"): "/nxentry/nxprocess",
                    "nxprocess": {
                        ("", "default"): "/nxentry/nxprocess/nxdata",
                        "nxdata": {
                            ("", "NX_class"): "NXdata",
                            ("", "signal"): "data",
                            "data": (1, 2, 3),
                        }
                    }
                }
            },
            hdf5_file)
        default = nxdata.get_default(hdf5_file)
        assert isinstance(default, nxdata.NXdata)
        assert default.group.name == "/nxentry/nxprocess/nxdata"

    def testRecursiveRelativePath(self, hdf5_file):
        dicttoh5(
            {
                ("", "default"): "nxentry",
                "nxentry": {
                    ("", "default"): "nxprocess",
                    "nxprocess": {
                        ("", "default"): "nxdata",
                        "nxdata": {
                            ("", "NX_class"): "NXdata",
                            ("", "signal"): "data",
                            "data": (1, 2, 3),
                        }
                    }
                }
            },
            hdf5_file)
        default = nxdata.get_default(hdf5_file)
        assert isinstance(default, nxdata.NXdata)
        assert default.group.name == "/nxentry/nxprocess/nxdata"

    def testLoop(self, hdf5_file):
        """Infinite loop of @default"""
        dicttoh5(
            {
                ("", "default"): "/nxentry",
                "nxentry": {
                    ("", "default"): "/nxentry",
                }
            },
            hdf5_file)
        default = nxdata.get_default(hdf5_file)
        assert default is None



# /*##########################################################################
#
# Copyright (c) 2015-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This package provides test of the root modules
"""

import logging


try:
    import pytest
except ImportError:
    logging.getLogger(__name__).error(
        "pytest is required to run the tests, please install it.")
    raise

def run_tests(module: str='silx', verbosity: int=0, args=()):
    """Run tests

    :param module: Name of the silx module to test (default: 'silx')
    :param verbosity: Requested level of verbosity
    :param args: List of extra arguments to pass to `pytest`
    """
    return pytest.main([
        '--pyargs',
        module,
        '--verbosity',
        str(verbosity),
        '-o python_files=["test/test*.py","test/Test*.py"]',
        '-o python_classes=["Test"]',
        '-o python_functions=["test"]',
    ] + list(args))



# /*##########################################################################
#
# Copyright (c) 2015-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides marching cubes implementation.

It provides a :class:`MarchingCubes` class allowing to build an isosurface
from data provided as a 3D data set or slice by slice.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "16/08/2017"


import numpy
cimport numpy as cnumpy
cimport cython

cimport silx.math.mc as mc


# From numpy_common.pxi to avoid warnings while compiling C code
# See this thread:
# https://mail.python.org/pipermail//cython-devel/2012-March/002137.html
cdef extern from *:
    bint FALSE "0"
    void import_array()
    void import_umath()

if FALSE:
    import_array()
    import_umath()


cdef class MarchingCubes:
    """Compute isosurface using marching cubes algorithm.

    It builds a surface from a 3D scalar dataset as a 3D contour at a
    given value.
    The resulting surface is not topologically correct.

    See: http://paulbourke.net/geometry/polygonise/

    Lorensen, W. E. and Cline, H. E. Marching cubes: A high resolution 3D
    surface construction algorithm. Computer Graphics, 21, 4 (July 1987).
    ACM, 163-169.

    Generated vertex and normal coordinates are in the same order
    as input array, i.e., (dim 0, dim 1, dim 2).

    Expected indices in memory of a (2, 2, 2) dataset:

           dim 0 (depth)
             |
             |
           4 +------+ 5
            /|     /|
           / |    / |
        6 +------+ 7|
          |  |   |  |
          |0 +---|--+ 1 --- dim 2 (width)
          | /    | /
          |/     |/
        2 +------+ 3
         /
        /
      dim 1 (height)

    Example with a 3D data set:

    >>> vertices, normals, indices = MarchingCubes(data, isolevel=1.)

    Example of code for processing a list of images:

    >>> mc = MarchingCubes(isolevel=1.)  # Create object with iso-level=1
    >>> previous_image = images[0]
    >>> for image in images[1:]:
    ...     mc.process_image(previous_image, image)  # Process one slice
    ...     previous_image = image

    >>> vertices = mc.get_vertices()  # Array of vertex positions
    >>> normals = mc.get_normals()  # Array of normals
    >>> triangle_indices = mc.get_indices()  # Array of indices of vertices

    :param data: 3D dataset of float32 or None
    :type data: numpy.ndarray of float32 of dimension 3
    :param float isolevel: The value for which to generate the isosurface
    :param bool invert_normals:
        True (default) for normals oriented in direction of gradient descent
    :param sampling: Sampling along each dimension (depth, height, width)
    """
    cdef mc.MarchingCubes[float, float] * c_mc  # Pointer to the C++ instance

    def __cinit__(self, data=None, isolevel=None,
                  invert_normals=True, sampling=(1, 1, 1)):
        self.c_mc = new mc.MarchingCubes[float, float](isolevel)
        self.c_mc.invert_normals = bool(invert_normals)
        self.c_mc.sampling[0] = sampling[0]
        self.c_mc.sampling[1] = sampling[1]
        self.c_mc.sampling[2] = sampling[2]

        if data is not None:
            self.process(data)

    def __dealloc__(self):
        del self.c_mc

    def __getitem__(self, key):
        """Allows one to unpack object as a single liner:

        vertices, normals, indices = MarchingCubes(...)
        """
        if key == 0:
            return self.get_vertices()
        elif key == 1:
            return self.get_normals()
        elif key == 2:
            return self.get_indices()
        else:
            raise IndexError("Index out of range")

    def process(self, data):
        """Compute an isosurface from a 3D scalar field.

        This builds vertices, normals and indices arrays.
        Vertices and normals coordinates are in the same order as input array,
        i.e., (dim 0, dim 1, dim 2).

        :param numpy.ndarray data: 3D scalar field
        """
        # Make sure data is a 3D contiguous array of native endian float32
        data = numpy.ascontiguousarray(data, dtype='=f4')
        assert data.ndim == 3
        cdef float[:] c_data = numpy.ravel(data)
        cdef unsigned int depth, height, width

        depth = data.shape[0]
        height = data.shape[1]
        width = data.shape[2]

        self.c_mc.process(&c_data[0], depth, height, width)

    def process_slice(self, slice0, slice1):
        """Process a new slice to build the isosurface.

        :param numpy.ndarray slice0: Slice previously provided as slice1.
        :param numpy.ndarray slice1: Slice to process.
        """
        # Make sure slices are 2D contiguous arrays of native endian float32
        slice0 = numpy.ascontiguousarray(slice0, dtype='=f4')
        assert slice0.ndim == 2
        slice1 = numpy.ascontiguousarray(slice1, dtype='=f4')
        assert slice1.ndim == 2

        assert slice0.shape[0] == slice1.shape[0]
        assert slice0.shape[1] == slice1.shape[1]

        cdef float[:] c_slice0 = numpy.ravel(slice0)
        cdef float[:] c_slice1 = numpy.ravel(slice1)

        if self.c_mc.depth == 0:
            # Starts a new isosurface, bootstrap with slice size
            self.c_mc.set_slice_size(slice1.shape[0], slice1.shape[1])

        assert slice1.shape[0] == self.c_mc.height
        assert slice1.shape[1] == self.c_mc.width

        self.c_mc.process_slice(&c_slice0[0], &c_slice1[0])

    def finish_process(self):
        """Clear internal cache after processing slice by slice."""
        self.c_mc.finish_process()

    def reset(self):
        """Reset internal resources including computed isosurface info."""
        self.c_mc.reset()

    @cython.embedsignature(False)
    @property
    def shape(self):
        """The shape of the processed scalar field (depth, height, width)."""
        return self.c_mc.depth, self.c_mc.height, self.c_mc.width

    @cython.embedsignature(False)
    @property
    def sampling(self):
        """The sampling over each dimension (depth, height, width).

        Default: 1, 1, 1
        """
        return (self.c_mc.sampling[0],
                self.c_mc.sampling[1],
                self.c_mc.sampling[2])

    @cython.embedsignature(False)
    @property
    def isolevel(self):
        """The iso-level at which to generate the isosurface"""
        return self.c_mc.isolevel

    @cython.embedsignature(False)
    @property
    def invert_normals(self):
        """True to use gradient descent as normals."""
        return self.c_mc.invert_normals

    def get_vertices(self):
        """Vertices currently computed (ndarray of dim NbVertices x 3)

        Order is dim0, dim1, dim2 (i.e., z, y, x if dim0 is depth).
        """
        return numpy.array(self.c_mc.vertices).reshape(-1, 3)

    def get_normals(self):
        """Normals currently computed (ndarray of dim NbVertices x 3)

        Order is dim0, dim1, dim2 (i.e., z, y, x if dim0 is depth).
        """
        return numpy.array(self.c_mc.normals).reshape(-1, 3)

    def get_indices(self):
        """Triangle indices currently computed (ndarray of dim NbTriangles x 3)
        """
        return numpy.array(self.c_mc.indices,
                           dtype=numpy.uint32).reshape(-1, 3)



# /*##########################################################################
#
# Copyright (c) 2004-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides an IPython console widget.

You can push variables - any python object - to the
console's interactive namespace. This provides users with an advanced way
of interacting with your program. For instance, if your program has a
:class:`PlotWidget` or a :class:`PlotWindow`, you can push a reference to
these widgets to allow your users to add curves, save data to files… by using
the widgets' methods from the console.

.. note::

    This module has a dependency on
    `qtconsole <https://pypi.org/project/qtconsole/>`_.
    An ``ImportError`` will be raised if it is
    imported while the dependencies are not satisfied.

Basic usage example::

    from silx.gui import qt
    from silx.gui.console import IPythonWidget

    app = qt.QApplication([])

    hello_button = qt.QPushButton("Hello World!", None)
    hello_button.show()

    console = IPythonWidget()
    console.show()
    console.pushVariables({"the_button": hello_button})

    app.exec()

This program will display a console widget and a push button in two separate
windows. You will be able to interact with the button from the console,
for example change its text::

    >>> the_button.setText("Spam spam")

An IPython interactive console is a powerful tool that enables you to work
with data and plot it.
See `this tutorial <https://plot.ly/python/ipython-notebook-tutorial/>`_
for more information on some of the rich features of IPython.
"""
__authors__ = ["Tim Rae", "V.A. Sole", "P. Knobel"]
__license__ = "MIT"
__date__ = "24/05/2016"

import logging

from . import qt

_logger = logging.getLogger(__name__)


# This widget cannot be used inside an interactive IPython shell.
# It would raise MultipleInstanceError("Multiple incompatible subclass
# instances of InProcessInteractiveShell are being created").
try:
    __IPYTHON__
except NameError:
    pass  # Not in IPython
else:
    msg = "Module " + __name__ + " cannot be used within an IPython shell"
    raise ImportError(msg)

try:
    from qtconsole.rich_jupyter_widget import RichJupyterWidget as \
        _RichJupyterWidget
except ImportError:
    try:
        from qtconsole.rich_ipython_widget import RichJupyterWidget as \
            _RichJupyterWidget
    except ImportError:
        from qtconsole.rich_ipython_widget import RichIPythonWidget as \
            _RichJupyterWidget

from qtconsole.inprocess import QtInProcessKernelManager

try:
    from ipykernel import version_info as _ipykernel_version_info
except ImportError:
    _ipykernel_version_info = None


class IPythonWidget(_RichJupyterWidget):
    """Live IPython console widget.

    .. image:: img/IPythonWidget.png

    :param custom_banner: Custom welcome message to be printed at the top of
       the console.
    """

    def __init__(self, parent=None, custom_banner=None, *args, **kwargs):
        if parent is not None:
            kwargs["parent"] = parent
        super(IPythonWidget, self).__init__(*args, **kwargs)
        if custom_banner is not None:
            self.banner = custom_banner
        self.setWindowTitle(self.banner)
        self.kernel_manager = kernel_manager = QtInProcessKernelManager()
        kernel_manager.start_kernel()

        # Monkey-patch to workaround issue:
        # https://github.com/ipython/ipykernel/issues/370
        if (_ipykernel_version_info is not None and
                _ipykernel_version_info[0] > 4 and
                _ipykernel_version_info[:3] <= (5, 1, 0)):
            def _abort_queues(*args, **kwargs):
                pass
            kernel_manager.kernel._abort_queues = _abort_queues

        self.kernel_client = kernel_client = self._kernel_manager.client()
        kernel_client.start_channels()

        def stop():
            kernel_client.stop_channels()
            kernel_manager.shutdown_kernel()
        self.exit_requested.connect(stop)

    def sizeHint(self):
        """Return a reasonable default size for usage in :class:`PlotWindow`"""
        return qt.QSize(500, 300)

    def pushVariables(self, variable_dict):
        """ Given a dictionary containing name / value pairs, push those
        variables to the IPython console widget.

        :param variable_dict: Dictionary of variables to be pushed to the
            console's interactive namespace (```{variable_name: object, …}```)
        """
        self.kernel_manager.kernel.shell.push(variable_dict)


class IPythonDockWidget(qt.QDockWidget):
    """Dock Widget including a :class:`IPythonWidget` inside
    a vertical layout.

    .. image:: img/IPythonDockWidget.png

    :param available_vars: Dictionary of variables to be pushed to the
        console's interactive namespace: ``{"variable_name": object, …}``
    :param custom_banner: Custom welcome message to be printed at the top of
        the console
    :param title: Dock widget title
    :param parent: Parent :class:`qt.QMainWindow` containing this
        :class:`qt.QDockWidget`
    """
    def __init__(self, parent=None, available_vars=None, custom_banner=None,
                 title="Console"):
        super(IPythonDockWidget, self).__init__(title, parent)

        self.ipyconsole = IPythonWidget(custom_banner=custom_banner)

        self.layout().setContentsMargins(0, 0, 0, 0)
        self.setWidget(self.ipyconsole)

        if available_vars is not None:
            self.ipyconsole.pushVariables(available_vars)


def main():
    """Run a Qt app with an IPython console"""
    app = qt.QApplication([])
    widget = IPythonDockWidget()
    widget.show()
    app.exec()


if __name__ == '__main__':
    main()



# /*##########################################################################
#
# Copyright (c) 2004-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Depracted module linking old PlotAction with the actions.xxx"""


__author__ = ["V.A. Sole", "T. Vincent"]
__license__ = "MIT"
__date__ = "01/06/2017"

from silx.utils.deprecation import deprecated_warning

deprecated_warning(type_='module',
                   name=__file__,
                   reason='PlotActions refactoring',
                   replacement='plot.actions',
                   since_version='0.6')

from .actions import PlotAction

from .actions.io import CopyAction
from .actions.io import PrintAction
from .actions.io import SaveAction

from .actions.control import ColormapAction
from .actions.control import CrosshairAction
from .actions.control import CurveStyleAction
from .actions.control import GridAction
from .actions.control import KeepAspectRatioAction
from .actions.control import PanWithArrowKeysAction
from .actions.control import ResetZoomAction
from .actions.control import XAxisAutoScaleAction
from .actions.control import XAxisLogarithmicAction
from .actions.control import YAxisAutoScaleAction
from .actions.control import YAxisLogarithmicAction
from .actions.control import YAxisInvertedAction
from .actions.control import ZoomInAction
from .actions.control import ZoomOutAction

from .actions.medfilt import MedianFilter1DAction
from .actions.medfilt import MedianFilter2DAction
from .actions.medfilt import MedianFilterAction

from .actions.histogram import PixelIntensitiesHistoAction

from .actions.fit import FitAction



# /*##########################################################################
#
# Copyright (c) 2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Test of async module."""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "09/03/2018"


import threading
import unittest


from concurrent.futures import wait
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from silx.gui.utils import concurrent


class TestSubmitToQtThread(TestCaseQt):
    """Test submission of tasks to Qt main thread"""

    def setUp(self):
        # Reset executor to test lazy-loading in different conditions
        concurrent._executor = None
        super(TestSubmitToQtThread, self).setUp()

    def _task(self, value1, value2):
        return value1, value2

    def _taskWithException(self, *args, **kwargs):
        raise RuntimeError('task exception')

    def testFromMainThread(self):
        """Call submitToQtMainThread from the main thread"""
        value1, value2 = 0, 1
        future = concurrent.submitToQtMainThread(self._task, value1, value2=value2)
        self.assertTrue(future.done())
        self.assertEqual(future.result(1), (value1, value2))
        self.assertIsNone(future.exception(1))

        future = concurrent.submitToQtMainThread(self._taskWithException)
        self.assertTrue(future.done())
        with self.assertRaises(RuntimeError):
            future.result(1)
        self.assertIsInstance(future.exception(1), RuntimeError)

    def _threadedTest(self):
        """Function run in a thread for the tests"""
        value1, value2 = 0, 1
        future = concurrent.submitToQtMainThread(self._task, value1, value2=value2)

        wait([future], 3)

        self.assertTrue(future.done())
        self.assertEqual(future.result(1), (value1, value2))
        self.assertIsNone(future.exception(1))

        future = concurrent.submitToQtMainThread(self._taskWithException)

        wait([future], 3)

        self.assertTrue(future.done())
        with self.assertRaises(RuntimeError):
            future.result(1)
        self.assertIsInstance(future.exception(1), RuntimeError)

    def testFromPythonThread(self):
        """Call submitToQtMainThread from a Python thread"""
        thread = threading.Thread(target=self._threadedTest)
        thread.start()
        for i in range(100):  # Loop over for 10 seconds
            self.qapp.processEvents()
            thread.join(0.1)
            if not thread.is_alive():
                break
        else:
            self.fail(('Thread task still running'))

    def testFromQtThread(self):
        """Call submitToQtMainThread from a Qt thread pool"""
        class Runner(qt.QRunnable):
            def __init__(self, fn):
                super(Runner, self).__init__()
                self._fn = fn

            def run(self):
                self._fn()

            def autoDelete(self):
                return True

        threadPool = qt.silxGlobalThreadPool()
        runner = Runner(self._threadedTest)
        threadPool.start(runner)
        for i in range(100):  # Loop over for 10 seconds
            self.qapp.processEvents()
            done = threadPool.waitForDone(100)
            if done:
                break
        else:
            self.fail('Thread pool task still running')



# /*##########################################################################
#
# Copyright (c) 2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ###########################################################################*/
"""Test SceneWidget"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "06/03/2019"


import pytest

import numpy

from silx.utils.testutils import ParametricTestCase
from silx.gui.utils.testutils import TestCaseQt
from silx.gui import qt

from silx.gui.plot3d.SceneWidget import SceneWidget


@pytest.mark.usefixtures("use_opengl")
class TestSceneWidget(TestCaseQt, ParametricTestCase):
    """Tests SceneWidget picking feature"""

    def setUp(self):
        super(TestSceneWidget, self).setUp()
        self.widget = SceneWidget()
        self.widget.show()
        self.qWaitForWindowExposed(self.widget)

    def tearDown(self):
        self.qapp.processEvents()
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        del self.widget
        super(TestSceneWidget, self).tearDown()

    def testFogEffect(self):
        """Test fog effect on scene primitive"""
        image = self.widget.addImage(numpy.arange(100).reshape(10, 10))
        scatter = self.widget.add3DScatter(*numpy.random.random(4000).reshape(4, -1))
        scatter.setTranslation(10, 10)
        scatter.setScale(10, 10, 10)

        self.widget.resetZoom('front')
        self.qapp.processEvents()

        self.widget.setFogMode(self.widget.FogMode.LINEAR)
        self.qapp.processEvents()

        self.widget.setFogMode(self.widget.FogMode.NONE)
        self.qapp.processEvents()



#!/usr/bin/env python
#
#    Project: S I L X project
#             https://github.com/silx-kit/silx
#
#    Copyright (C) 2012-2021 European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#

__author__ = "Jerome Kieffer"
__contact__ = "Jerome.Kieffer@ESRF.eu"
__license__ = "MIT"
__copyright__ = "2012-2017 European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "29/09/2021"
__status__ = "stable"
__all__ = ["ocl", "pyopencl", "mf", "release_cl_buffers", "allocate_cl_buffers",
           "measure_workgroup_size", "kernel_workgroup_size"]

import os
import logging

import numpy

from .utils import get_opencl_code

logger = logging.getLogger(__name__)

if os.environ.get("SILX_OPENCL") in ["0", "False"]:
    logger.info("Use of OpenCL has been disabled from environment variable: SILX_OPENCL=0")
    pyopencl = None
else:
    try:
        import pyopencl
    except ImportError:
        logger.warning("Unable to import pyOpenCl. Please install it from: https://pypi.org/project/pyopencl")
        pyopencl = None
    else:
        try:
            pyopencl.get_platforms()
        except pyopencl.LogicError:
            logger.warning("The module pyOpenCL has been imported but can't be used here")
            pyopencl = None
        else:
            import pyopencl.array as array
            mf = pyopencl.mem_flags

if pyopencl is None:

    # Define default mem flags
    class mf(object):
        WRITE_ONLY = 1
        READ_ONLY = 1
        READ_WRITE = 1

FLOP_PER_CORE = {"GPU": 64,  # GPU, Fermi at least perform 64 flops per cycle/multicore, G80 were at 24 or 48 ...
                 "CPU": 4,  # CPU, at least intel's have 4 operation per cycle
                 "ACC": 8}  # ACC: the Xeon-phi (MIC) appears to be able to process 8 Flops per hyperthreaded-core

# Sources : https://en.wikipedia.org/wiki/CUDA
NVIDIA_FLOP_PER_CORE = {(1, 0): 24,   # Guessed !
                        (1, 1): 24,   # Measured on G98 [Quadro NVS 295]
                        (1, 2): 24,   # Guessed !
                        (1, 3): 24,   # measured on a GT285 (GT200)
                        (2, 0): 64,   # Measured on a 580 (GF110)
                        (2, 1): 96,   # Measured on Quadro2000 GF106GL
                        (3, 0): 384,  # Guessed!
                        (3, 5): 384,  # Measured on K20
                        (3, 7): 384,  # K80: Guessed!
                        (5, 0): 256,  # Maxwell 4 warps/SM 2 flops/ CU
                        (5, 2): 256,  # Titan-X
                        (5, 3): 256,  # TX1
                        (6, 0): 128,  # GP100
                        (6, 1): 128,  # GP104
                        (6, 2): 128,  # ?
                        (7, 0): 128,  # Volta  # measured on Telsa V100
                        (7, 2): 128,  # Volta  ? 
                        (7, 5): 128,  # Turing # measured on RTX 6000
                        (8, 0): 128,  # Ampere # measured on Tesla A100
                        (8, 6): 256,  # Ampere # measured on RTX A5000
                        }

AMD_FLOP_PER_CORE = 160  # Measured on a M7820 10 core, 700MHz 1120GFlops


class Device(object):
    """
    Simple class that contains the structure of an OpenCL device
    """

    def __init__(self, name="None", dtype=None, version=None, driver_version=None,
                 extensions="", memory=None, available=None,
                 cores=None, frequency=None, flop_core=None, idx=0, workgroup=1):
        """
        Simple container with some important data for the OpenCL device description.

        :param name: name of the device
        :param dtype: device type: CPU/GPU/ACC...
        :param version: driver version
        :param driver_version:
        :param extensions: List of opencl extensions
        :param memory: maximum memory available on the device
        :param available: is the device deactivated or not
        :param cores: number of SM/cores
        :param frequency: frequency of the device
        :param flop_core: Flopating Point operation per core per cycle
        :param idx: index of the device within the platform
        :param workgroup: max workgroup size
        """
        self.name = name.strip()
        self.type = dtype
        self.version = version
        self.driver_version = driver_version
        self.extensions = extensions.split()
        self.memory = memory
        self.available = available
        self.cores = cores
        self.frequency = frequency
        self.id = idx
        self.max_work_group_size = workgroup
        if not flop_core:
            flop_core = FLOP_PER_CORE.get(dtype, 1)
        if cores and frequency:
            self.flops = cores * frequency * flop_core
        else:
            self.flops = flop_core

    def __repr__(self):
        return "%s" % self.name

    def pretty_print(self):
        """
        Complete device description

        :return: string
        """
        lst = ["Name\t\t:\t%s" % self.name,
               "Type\t\t:\t%s" % self.type,
               "Memory\t\t:\t%.3f MB" % (self.memory / 2.0 ** 20),
               "Cores\t\t:\t%s CU" % self.cores,
               "Frequency\t:\t%s MHz" % self.frequency,
               "Speed\t\t:\t%.3f GFLOPS" % (self.flops / 1000.),
               "Version\t\t:\t%s" % self.version,
               "Available\t:\t%s" % self.available]
        return os.linesep.join(lst)

    def set_unavailable(self):
        """Use this method to flag a faulty device
        """
        self.available = False


class Platform(object):
    """
    Simple class that contains the structure of an OpenCL platform
    """

    def __init__(self, name="None", vendor="None", version=None, extensions=None, idx=0):
        """
        Class containing all descriptions of a platform and all devices description within that platform.

        :param name: platform name
        :param vendor: name of the brand/vendor
        :param version:
        :param extensions: list of the extension provided by the platform to all of its devices
        :param idx: index of the platform
        """
        self.name = name.strip()
        self.vendor = vendor.strip()
        self.version = version
        self.extensions = extensions.split()
        self.devices = []
        self.id = idx

    def __repr__(self):
        return "%s" % self.name

    def add_device(self, device):
        """
        Add new device to the platform

        :param device: Device instance
        """
        self.devices.append(device)

    def get_device(self, key):
        """
        Return a device according to key

        :param key: identifier for a device, either it's id (int) or it's name
        :type key: int or str
        """
        out = None
        try:
            devid = int(key)
        except ValueError:
            for a_dev in self.devices:
                if a_dev.name == key:
                    out = a_dev
        else:
            if len(self.devices) > devid > 0:
                out = self.devices[devid]
        return out


def _measure_workgroup_size(device_or_context, fast=False):
    """Mesure the maximal work group size of the given device

    DEPRECATED since not perfectly correct !

    :param device_or_context: instance of pyopencl.Device or pyopencl.Context
                    or 2-tuple (platformid,deviceid)
    :param fast: ask the kernel the valid value, don't probe it
    :return: maximum size for the workgroup
    """
    if isinstance(device_or_context, pyopencl.Device):
        try:
            ctx = pyopencl.Context(devices=[device_or_context])
        except pyopencl._cl.LogicError as error:
            platform = device_or_context.platform
            platformid = pyopencl.get_platforms().index(platform)
            deviceid = platform.get_devices().index(device_or_context)
            ocl.platforms[platformid].devices[deviceid].set_unavailable()
            raise RuntimeError("Unable to create context on %s/%s: %s" % (platform, device_or_context, error))
        else:
            device = device_or_context
    elif isinstance(device_or_context, pyopencl.Context):
        ctx = device_or_context
        device = device_or_context.devices[0]
    elif isinstance(device_or_context, (tuple, list)) and len(device_or_context) == 2:
        ctx = ocl.create_context(platformid=device_or_context[0],
                                 deviceid=device_or_context[1])
        device = ctx.devices[0]
    else:
        raise RuntimeError("""given parameter device_or_context is not an
            instanciation of a device or a context""")
    shape = device.max_work_group_size
    # get the context

    assert ctx is not None
    queue = pyopencl.CommandQueue(ctx)

    max_valid_wg = 1
    data = numpy.random.random(shape).astype(numpy.float32)
    d_data = pyopencl.array.to_device(queue, data)
    d_data_1 = pyopencl.array.empty_like(d_data)
    d_data_1.fill(numpy.float32(1.0))

    program = pyopencl.Program(ctx, get_opencl_code("addition")).build()
    if fast:
        max_valid_wg = program.addition.get_work_group_info(pyopencl.kernel_work_group_info.WORK_GROUP_SIZE, device)
    else:
        maxi = int(round(numpy.log2(shape)))
        for i in range(maxi + 1):
            d_res = pyopencl.array.empty_like(d_data)
            wg = 1 << i
            try:
                evt = program.addition(
                    queue, (shape,), (wg,),
                    d_data.data, d_data_1.data, d_res.data, numpy.int32(shape))
                evt.wait()
            except Exception as error:
                logger.info("%s on device %s for WG=%s/%s", error, device.name, wg, shape)
                program = queue = d_res = d_data_1 = d_data = None
                break
            else:
                res = d_res.get()
                good = numpy.allclose(res, data + 1)
                if good:
                    if wg > max_valid_wg:
                        max_valid_wg = wg
                else:
                    logger.warning("ArithmeticError on %s for WG=%s/%s", wg, device.name, shape)

    return max_valid_wg


def _is_nvidia_gpu(vendor, devtype):
    return (vendor == "NVIDIA Corporation") and (devtype == "GPU")


class OpenCL(object):
    """
    Simple class that wraps the structure ocl_tools_extended.h

    This is a static class.
    ocl should be the only instance and shared among all python modules.
    """

    platforms = []
    nb_devices = 0
    context_cache = {}  # key: 2-tuple of int, value: context
    if pyopencl:
        platform = device = pypl = devtype = extensions = pydev = None
        for idx, platform in enumerate(pyopencl.get_platforms()):
            pypl = Platform(platform.name, platform.vendor, platform.version, platform.extensions, idx)
            for idd, device in enumerate(platform.get_devices()):
                ####################################################
                # Nvidia does not report int64 atomics (we are using) ...
                # this is a hack around as any nvidia GPU with double-precision supports int64 atomics
                ####################################################
                extensions = device.extensions
                if (pypl.vendor == "NVIDIA Corporation") and ('cl_khr_fp64' in extensions):
                    extensions += ' cl_khr_int64_base_atomics cl_khr_int64_extended_atomics'
                try:
                    devtype = pyopencl.device_type.to_string(device.type).upper()
                except ValueError:
                    # pocl does not describe itself as a CPU !
                    devtype = "CPU"
                if len(devtype) > 3:
                    if "GPU" in devtype:
                        devtype = "GPU"
                    elif "ACC" in devtype:
                        devtype = "ACC"
                    elif "CPU" in devtype:
                        devtype = "CPU"
                    else:
                        devtype = devtype[:3]
                if _is_nvidia_gpu(device.vendor, devtype) and ("compute_capability_major_nv" in dir(device)):
                    try:
                        comput_cap = device.compute_capability_major_nv, device.compute_capability_minor_nv
                    except pyopencl.LogicError:
                        flop_core = FLOP_PER_CORE["GPU"]
                    else:
                        flop_core = NVIDIA_FLOP_PER_CORE.get(comput_cap, FLOP_PER_CORE["GPU"])
                elif (pypl.vendor == "Advanced Micro Devices, Inc.") and (devtype == "GPU"):
                    flop_core = AMD_FLOP_PER_CORE
                elif devtype == "CPU":
                    flop_core = FLOP_PER_CORE.get(devtype, 1)
                else:
                    flop_core = 1
                workgroup = device.max_work_group_size
                if (devtype == "CPU") and (pypl.vendor == "Apple"):
                    logger.info("For Apple's OpenCL on CPU: Measuring actual valid max_work_goup_size.")
                    workgroup = _measure_workgroup_size(device, fast=True)
                if (devtype == "GPU") and os.environ.get("GPU") == "False":
                    # Environment variable to disable GPU devices
                    continue
                pydev = Device(device.name, devtype, device.version, device.driver_version, extensions,
                               device.global_mem_size, bool(device.available), device.max_compute_units,
                               device.max_clock_frequency, flop_core, idd, workgroup)
                pypl.add_device(pydev)
                nb_devices += 1
            platforms.append(pypl)
        del platform, device, pypl, devtype, extensions, pydev

    def __repr__(self):
        out = ["OpenCL devices:"]
        for platformid, platform in enumerate(self.platforms):
            deviceids = ["(%s,%s) %s" % (platformid, deviceid, dev.name)
                         for deviceid, dev in enumerate(platform.devices)]
            out.append("[%s] %s: " % (platformid, platform.name) + ", ".join(deviceids))
        return os.linesep.join(out)

    def get_platform(self, key):
        """
        Return a platform according

        :param key: identifier for a platform, either an Id (int) or it's name
        :type key: int or str
        """
        out = None
        try:
            platid = int(key)
        except ValueError:
            for a_plat in self.platforms:
                if a_plat.name == key:
                    out = a_plat
        else:
            if len(self.platforms) > platid > 0:
                out = self.platforms[platid]
        return out

    def select_device(self, dtype="ALL", memory=None, extensions=None, best=True, **kwargs):
        """
        Select a device based on few parameters (at the end, keep the one with most memory)

        :param dtype: "gpu" or "cpu" or "all" ....
        :param memory: minimum amount of memory (int)
        :param extensions: list of extensions to be present
        :param best: shall we look for the
        :returns: A tuple of plateform ID and device ID, else None if nothing
            found
        """
        if extensions is None:
            extensions = []
        if "type" in kwargs:
            dtype = kwargs["type"].upper()
        else:
            dtype = dtype.upper()
        if len(dtype) > 3:
            dtype = dtype[:3]
        best_found = None
        for platformid, platform in enumerate(self.platforms):
            for deviceid, device in enumerate(platform.devices):
                if not device.available:
                    continue
                if (dtype in ["ALL", "DEF"]) or (device.type == dtype):
                    if (memory is None) or (memory <= device.memory):
                        found = True
                        for ext in extensions:
                            if ext not in device.extensions:
                                found = False
                        if found:
                            if not best:
                                return platformid, deviceid
                            else:
                                if not best_found:
                                    best_found = platformid, deviceid, device.flops
                                elif best_found[2] < device.flops:
                                    best_found = platformid, deviceid, device.flops
        if best_found:
            return best_found[0], best_found[1]

        # Nothing found
        return None

    def create_context(self, devicetype="ALL", useFp64=False, platformid=None,
                       deviceid=None, cached=True, memory=None, extensions=None):
        """
        Choose a device and initiate a context.

        Devicetypes can be GPU,gpu,CPU,cpu,DEF,ACC,ALL.
        Suggested are GPU,CPU.
        For each setting to work there must be such an OpenCL device and properly installed.
        E.g.: If Nvidia driver is installed, GPU will succeed but CPU will fail.
              The AMD SDK kit is required for CPU via OpenCL.
        :param devicetype: string in ["cpu","gpu", "all", "acc"]
        :param useFp64: boolean specifying if double precision will be used: deprecated use extensions=["cl_khr_fp64"]
        :param platformid: integer
        :param deviceid: integer
        :param cached: True if we want to cache the context
        :param memory: minimum amount of memory of the device
        :param extensions: list of extensions to be present
        :return: OpenCL context on the selected device
        """
        if extensions is None:
            extensions = []
        if useFp64:
            logger.warning("Deprecation: please select your device using the extension name!, i.e. extensions=['cl_khr_fp64']")
            extensions.append('cl_khr_fp64')

        if (platformid is not None) and (deviceid is not None):
            platformid = int(platformid)
            deviceid = int(deviceid)
        elif "PYOPENCL_CTX" in os.environ:
            pyopencl_ctx = [int(i) if i.isdigit() else 0 for i in os.environ["PYOPENCL_CTX"].split(":")]
            pyopencl_ctx += [0] * (2 - len(pyopencl_ctx))  # pad with 0
            platformid, deviceid = pyopencl_ctx
        else:
            ids = ocl.select_device(type=devicetype, extensions=extensions)
            if ids:
                platformid, deviceid = ids
        ctx = None
        if (platformid is not None) and (deviceid is not None):
            if (platformid, deviceid) in self.context_cache:
                ctx = self.context_cache[(platformid, deviceid)]
            else:
                try:
                    ctx = pyopencl.Context(devices=[pyopencl.get_platforms()[platformid].get_devices()[deviceid]])
                except pyopencl._cl.LogicError as error:
                    self.platforms[platformid].devices[deviceid].set_unavailable()
                    logger.warning("Unable to create context on %s/%s: %s", platformid, deviceid, error)
                    ctx = None
                else:
                    if cached:
                        self.context_cache[(platformid, deviceid)] = ctx
        if ctx is None:
            logger.warning("Last chance to get an OpenCL device ... probably not the one requested")
            ctx = pyopencl.create_some_context(interactive=False)
        return ctx

    def device_from_context(self, context):
        """
        Retrieves the Device from the context

        :param context: OpenCL context
        :return: instance of Device
        """
        odevice = context.devices[0]
        oplat = odevice.platform
        device_id = oplat.get_devices().index(odevice)
        platform_id = pyopencl.get_platforms().index(oplat)
        return self.platforms[platform_id].devices[device_id]


if pyopencl:
    ocl = OpenCL()
    if ocl.nb_devices == 0:
        ocl = None
else:
    ocl = None


def release_cl_buffers(cl_buffers):
    """
    :param cl_buffers: the buffer you want to release
    :type cl_buffers: dict(str, pyopencl.Buffer)

    This method release the memory of the buffers store in the dict
    """
    for key, buffer_ in cl_buffers.items():
        if buffer_ is not None:
            if isinstance(buffer_, pyopencl.array.Array):
                try:
                    buffer_.data.release()
                except pyopencl.LogicError:
                    logger.error("Error while freeing buffer %s", key)
            else:
                try:
                    buffer_.release()
                except pyopencl.LogicError:
                    logger.error("Error while freeing buffer %s", key)
            cl_buffers[key] = None
    return cl_buffers


def allocate_cl_buffers(buffers, device=None, context=None):
    """
    :param buffers: the buffers info use to create the pyopencl.Buffer
    :type buffers: list(std, flag, numpy.dtype, int)
    :param device: one of the context device
    :param context: opencl contextdevice
    :return: a dict containing the instanciated pyopencl.Buffer
    :rtype: dict(str, pyopencl.Buffer)

    This method instanciate the pyopencl.Buffer from the buffers
    description.
    """
    mem = {}
    if device is None:
        device = ocl.device_from_context(context)

    # check if enough memory is available on the device
    ualloc = 0
    for _, _, dtype, size in buffers:
        ualloc += numpy.dtype(dtype).itemsize * size
    memory = device.memory
    logger.info("%.3fMB are needed on device which has %.3fMB",
                ualloc / 1.0e6, memory / 1.0e6)
    if ualloc >= memory:
        memError = "Fatal error in allocate_buffers."
        memError += "Not enough device memory for buffers"
        memError += "(%lu requested, %lu available)" % (ualloc, memory)
        raise MemoryError(memError)  # noqa

    # do the allocation
    try:
        for name, flag, dtype, size in buffers:
            mem[name] = pyopencl.Buffer(context, flag,
                                        numpy.dtype(dtype).itemsize * size)
    except pyopencl.MemoryError as error:
        release_cl_buffers(mem)
        raise MemoryError(error)

    return mem


def allocate_texture(ctx, shape, hostbuf=None, support_1D=False):
    """
    Allocate an OpenCL image ("texture").

    :param ctx: OpenCL context
    :param shape: Shape of the image. Note that pyopencl and OpenCL < 1.2
        do not support 1D images, so 1D images are handled as 2D with one row
    :param support_1D: force the image to be 1D if the shape has only one dim
    """
    if len(shape) == 1 and not(support_1D):
        shape = (1,) + shape
    return pyopencl.Image(
        ctx,
        pyopencl.mem_flags.READ_ONLY | pyopencl.mem_flags.USE_HOST_PTR,
        pyopencl.ImageFormat(
            pyopencl.channel_order.INTENSITY,
            pyopencl.channel_type.FLOAT
        ),
        hostbuf=numpy.zeros(shape[::-1], dtype=numpy.float32)
    )


def check_textures_availability(ctx):
    """
    Check whether textures are supported on the current OpenCL context.

    :param ctx: OpenCL context
    """
    try:
        dummy_texture = allocate_texture(ctx, (16, 16))
        # Need to further access some attributes (pocl)
        dummy_height = dummy_texture.height
        textures_available = True
        del dummy_texture, dummy_height
    except (pyopencl.RuntimeError, pyopencl.LogicError):
        textures_available = False
    # Nvidia Fermi GPUs (compute capability 2.X) do not support opencl read_imagef
    # There is no way to detect this until a kernel is compiled
    try:
        cc = ctx.devices[0].compute_capability_major_nv
        textures_available &= (cc >= 3)
    except (pyopencl.LogicError, AttributeError):  # probably not a Nvidia GPU
        pass
    #
    return textures_available


def measure_workgroup_size(device):
    """Measure the actual size of the workgroup

    :param device: device or context or 2-tuple with indexes
    :return: the actual measured workgroup size

    if device is "all", returns a dict with all devices with their ids as keys.
    """
    if (ocl is None) or (device is None):
        return None

    if isinstance(device, tuple) and (len(device) == 2):
        # this is probably a tuple (platformid, deviceid)
        device = ocl.create_context(platformid=device[0], deviceid=device[1])

    if device == "all":
        res = {}
        for pid, platform in enumerate(ocl.platforms):
            for did, _devices in enumerate(platform.devices):
                tup = (pid, did)
                res[tup] = measure_workgroup_size(tup)
    else:
        res = _measure_workgroup_size(device)
    return res


def query_kernel_info(program, kernel, what="WORK_GROUP_SIZE"):
    """Extract the compile time information from a kernel

    :param program: OpenCL program
    :param kernel: kernel or name of the kernel
    :param what: what is the query about ?
    :return: int or 3-int for the workgroup size.
    
    Possible information available are:
    * 'COMPILE_WORK_GROUP_SIZE': Returns the work-group size specified inside the kernel (__attribute__((reqd_work_gr oup_size(X, Y, Z))))
    * 'GLOBAL_WORK_SIZE': maximum global size that can be used to execute a kernel  #OCL2.1!
    * 'LOCAL_MEM_SIZE': amount of local memory in bytes being used by the kernel
    * 'PREFERRED_WORK_GROUP_SIZE_MULTIPLE': preferred multiple of workgroup size for launch. This is a performance hint.
    * 'PRIVATE_MEM_SIZE' Returns the minimum amount of private memory, in bytes, used by each workitem in the kernel
    * 'WORK_GROUP_SIZE': maximum work-group size that can be used to execute a kernel on a specific device given by device
    
    Further information on:
    https://www.khronos.org/registry/OpenCL/sdk/1.1/docs/man/xhtml/clGetKernelWorkGroupInfo.html
    
    """
    assert isinstance(program, pyopencl.Program)
    if not isinstance(kernel, pyopencl.Kernel):
        kernel_name = kernel
        assert kernel in (k.function_name for k in program.all_kernels()), "the kernel exists"
        kernel = program.__getattr__(kernel_name)

    device = program.devices[0]
    query_wg = getattr(pyopencl.kernel_work_group_info, what)
    return kernel.get_work_group_info(query_wg, device)


def kernel_workgroup_size(program, kernel):
    """Extract the compile time maximum workgroup size

    :param program: OpenCL program
    :param kernel: kernel or name of the kernel
    :return: the maximum acceptable workgroup size for the given kernel
    """
    return query_kernel_info(program, kernel, what="WORK_GROUP_SIZE")



# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
Project: Sift implementation in Python + OpenCL
         https://github.com/silx-kit/silx
"""

__authors__ = ["Jérôme Kieffer", "Pierre Paleo"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "06/09/2017"
__status__ = "Production"

import os
import numpy
from .. import resources
from math import log, ceil


def calc_size(shape, blocksize):
    """
    Calculate the optimal size for a kernel according to the workgroup size
    """
    if "__len__" in dir(blocksize):
        return tuple((int(i) + int(j) - 1) & ~(int(j) - 1) for i, j in zip(shape, blocksize))
    else:
        return tuple((int(i) + int(blocksize) - 1) & ~(int(blocksize) - 1) for i in shape)


def nextpower(n):
    """Calculate the power of two

    :param n: an integer, for example 100
    :return: another integer, 100-> 128
    """
    return 1 << int(ceil(log(n, 2)))


def sizeof(shape, dtype="uint8"):
    """
    Calculate the number of bytes needed to allocate for a given structure

    :param shape: size or tuple of sizes
    :param dtype: data type
    """
    itemsize = numpy.dtype(dtype).itemsize
    cnt = 1
    if "__len__" in dir(shape):
        for dim in shape:
            cnt *= dim
    else:
        cnt = int(shape)
    return cnt * itemsize


def get_cl_file(resource):
    """get the full path of a openCL resource file

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx".
    See also :func:`silx.resources.register_resource_directory`.

    :param str resource: Resource name. File name contained if the `opencl`
        directory of the resources.
    :return: the full path of the openCL source file
    """
    if not resource.endswith(".cl"):
        resource += ".cl"
    return resources._resource_filename(resource,
                                        default_directory="opencl")


def read_cl_file(filename):
    """
    :param filename: read an OpenCL file and apply a preprocessor
    :return: preprocessed source code
    """
    with open(get_cl_file(filename), "r") as f:
        # Dummy preprocessor which removes the #include
        lines = [i for i in f.readlines() if not i.startswith("#include ")]
    return "".join(lines)


get_opencl_code = read_cl_file


def concatenate_cl_kernel(filenames):
    """Concatenates all the kernel from the list of files

    :param filenames: filenames containing the kernels
    :type filenames: list of str which can be filename of kernel as a string.
    :return: a string with all kernels concatenated

    this method concatenates all the kernel from the list
    """
    return os.linesep.join(read_cl_file(fn) for fn in filenames)




class ConvolutionInfos(object):
    allowed_axes = {
        "1D": [None],
        "separable_2D_1D_2D": [None, (0, 1), (1, 0)],
        "batched_1D_2D": [(0,), (1,)],
        "separable_3D_1D_3D": [
            None,
            (0, 1, 2),
            (1, 2, 0),
            (2, 0, 1),
            (2, 1, 0),
            (1, 0, 2),
            (0, 2, 1)
        ],
        "batched_1D_3D": [(0,), (1,), (2,)],
        "batched_separable_2D_1D_3D": [(0,), (1,), (2,)], # unsupported (?)
        "2D": [None],
        "batched_2D_3D": [(0,), (1,), (2,)],
        "separable_3D_2D_3D": [
            (1, 0),
            (0, 1),
            (2, 0),
            (0, 2),
            (1, 2),
            (2, 1),
        ],
        "3D": [None],
    }
    use_cases = {
        (1, 1): {
            "1D": {
                "name": "1D convolution on 1D data",
                "kernels": ["convol_1D_X"],
            },
        },
        (2, 2): {
            "2D": {
                "name": "2D convolution on 2D data",
                "kernels": ["convol_2D_XY"],
            },
        },
        (3, 3): {
            "3D": {
                "name": "3D convolution on 3D data",
                "kernels": ["convol_3D_XYZ"],
            },
        },
        (2, 1): {
            "separable_2D_1D_2D": {
                "name": "Separable (2D->1D) convolution on 2D data",
                "kernels": ["convol_1D_X", "convol_1D_Y"],
            },
            "batched_1D_2D": {
                "name": "Batched 1D convolution on 2D data",
                "kernels": ["convol_1D_X", "convol_1D_Y"],
            },
        },
        (3, 1): {
            "separable_3D_1D_3D": {
                "name": "Separable (3D->1D) convolution on 3D data",
                "kernels": ["convol_1D_X", "convol_1D_Y", "convol_1D_Z"],
            },
            "batched_1D_3D": {
                "name": "Batched 1D convolution on 3D data",
                "kernels": ["convol_1D_X", "convol_1D_Y", "convol_1D_Z"],
            },
            "batched_separable_2D_1D_3D": {
                "name": "Batched separable (2D->1D) convolution on 3D data",
                "kernels": ["convol_1D_X", "convol_1D_Y", "convol_1D_Z"],
            },
        },
        (3, 2): {
            "separable_3D_2D_3D": {
                "name": "Separable (3D->2D) convolution on 3D data",
                "kernels": ["convol_2D_XY", "convol_2D_XZ", "convol_2D_YZ"],
            },
            "batched_2D_3D": {
                "name": "Batched 2D convolution on 3D data",
                "kernels": ["convol_2D_XY", "convol_2D_XZ", "convol_2D_YZ"],
            },
        },
    }







